import copy
from dataclasses import dataclass

import vyper.utils as util
from vyper.codegen.abi_encoder import abi_encode
from vyper.codegen.core import (
    _freshname,
    add_ofst,
    calculate_type_for_external_return,
    check_assign,
    check_external_call,
    dummy_node_for_type,
    eval_once_check,
    get_type_for_exact_size,
    make_setter,
    needs_clamp,
    unwrap_location,
    wrap_value_for_external_return,
)
from vyper.codegen.ir_node import Encoding, IRnode
from vyper.evm.address_space import MEMORY
from vyper.exceptions import TypeCheckFailure
from vyper.semantics.types import InterfaceT, TupleT
from vyper.semantics.types.function import StateMutability


@dataclass
class _CallKwargs:
    value: IRnode
    gas: IRnode
    skip_contract_check: bool
    default_return_value: IRnode


def _pack_arguments(fn_type, args, context):
    # abi encoding just treats all args as a big tuple
    args_tuple_t = TupleT([x.typ for x in args])
    args_as_tuple = IRnode.from_list(["multi"] + [x for x in args], typ=args_tuple_t)
    args_abi_t = args_tuple_t.abi_type

    # sanity typecheck - make sure the arguments can be assigned
    dst_tuple_t = TupleT(fn_type.argument_types[: len(args)])
    check_assign(dummy_node_for_type(dst_tuple_t), args_as_tuple)

    if fn_type.return_type is not None:
        return_abi_t = calculate_type_for_external_return(fn_type.return_type).abi_type

        # we use the same buffer for args and returndata,
        # so allocate enough space here for the returndata too.
        buflen = max(args_abi_t.size_bound(), return_abi_t.size_bound())
    else:
        buflen = args_abi_t.size_bound()

    buflen += 32  # padding for the method id

    buf_t = get_type_for_exact_size(buflen)
    buf = context.new_internal_variable(buf_t)

    args_ofst = add_ofst(buf, 28)
    args_len = args_abi_t.size_bound() + 4

    abi_signature = fn_type.name + dst_tuple_t.abi_type.selector_name()

    # layout:
    # 32 bytes                 | args
    # 0x..00<method_id_4bytes> | args
    # the reason for the left padding is just so the alignment is easier.
    # XXX: we could align to buf (and also keep code size small) by using
    # (mstore buf (shl signature.method_id 224))
    pack_args = ["seq"]
    pack_args.append(["mstore", buf, util.method_id_int(abi_signature)])

    if len(args) != 0:
        pack_args.append(abi_encode(add_ofst(buf, 32), args_as_tuple, context, bufsz=buflen))

    return buf, pack_args, args_ofst, args_len


def _unpack_returndata(buf, fn_type, call_kwargs, contract_address, context, expr):
    return_t = fn_type.return_type

    if return_t is None:
        return ["pass"], 0, 0

    wrapped_return_t = calculate_type_for_external_return(return_t)

    abi_return_t = wrapped_return_t.abi_type

    min_return_size = abi_return_t.min_size()
    max_return_size = abi_return_t.size_bound()
    assert 0 < min_return_size <= max_return_size

    ret_ofst = buf
    ret_len = max_return_size

    encoding = Encoding.ABI

    assert buf.location == MEMORY
    buf = copy.copy(buf)
    buf.typ = wrapped_return_t
    buf.encoding = encoding
    buf.annotation = f"{expr.node_source_code} returndata buffer"

    unpacker = ["seq"]

    # revert when returndatasize is not in bounds
    # (except when return_override is provided.)
    if not call_kwargs.skip_contract_check:
        assertion = IRnode.from_list(
            ["assert", ["ge", "returndatasize", min_return_size]],
            error_msg="returndatasize too small",
        )
        unpacker.append(assertion)

    assert isinstance(wrapped_return_t, TupleT)

    # unpack strictly
    if needs_clamp(wrapped_return_t, encoding):
        return_buf = context.new_internal_variable(wrapped_return_t)

        # note: make_setter does ABI decoding and clamps
        unpacker.append(make_setter(return_buf, buf))
    else:
        return_buf = buf

    if call_kwargs.default_return_value is not None:
        # if returndatasize == 0:
        #    copy return override to buf
        # else:
        #    do the other stuff

        override_value = wrap_value_for_external_return(call_kwargs.default_return_value)
        stomp_return_buffer = ["seq"]
        if not call_kwargs.skip_contract_check:
            stomp_return_buffer.append(_extcodesize_check(contract_address))
        stomp_return_buffer.append(make_setter(return_buf, override_value))
        unpacker = ["if", ["eq", "returndatasize", 0], stomp_return_buffer, unpacker]

    unpacker = ["seq", unpacker, return_buf]

    return unpacker, ret_ofst, ret_len


def _parse_kwargs(call_expr, context):
    from vyper.codegen.expr import Expr  # TODO rethink this circular import

    def _bool(x):
        assert x.value in (0, 1), "type checker missed this"
        return bool(x.value)

    # note: codegen for kwarg values in AST order
    call_kwargs = {kw.arg: Expr(kw.value, context).ir_node for kw in call_expr.keywords}

    ret = _CallKwargs(
        value=unwrap_location(call_kwargs.pop("value", IRnode(0))),
        gas=unwrap_location(call_kwargs.pop("gas", IRnode("gas"))),
        skip_contract_check=_bool(call_kwargs.pop("skip_contract_check", IRnode(0))),
        default_return_value=call_kwargs.pop("default_return_value", None),
    )

    if len(call_kwargs) != 0:  # pragma: nocover
        raise TypeCheckFailure(f"Unexpected keyword arguments: {call_kwargs}")

    return ret


def _extcodesize_check(address):
    return IRnode.from_list(["assert", ["extcodesize", address]], error_msg="extcodesize is zero")


def _external_call_helper(contract_address, args_ir, call_kwargs, call_expr, context):
    fn_type = call_expr.func._metadata["type"]

    # sanity check
    assert fn_type.n_positional_args <= len(args_ir) <= fn_type.n_total_args

    ret = ["seq"]

    # this is a sanity check to prevent double evaluation of the external call
    # in the codegen pipeline. if the external call gets doubly evaluated,
    # a duplicate label exception will get thrown during assembly.
    ret.append(eval_once_check(_freshname(call_expr.node_source_code)))

    buf, arg_packer, args_ofst, args_len = _pack_arguments(fn_type, args_ir, context)

    ret_unpacker, ret_ofst, ret_len = _unpack_returndata(
        buf, fn_type, call_kwargs, contract_address, context, call_expr
    )

    ret += arg_packer

    if fn_type.return_type is None and not call_kwargs.skip_contract_check:
        # if we do not expect return data, check that a contract exists at the
        # target address. we must perform this check BEFORE the call because
        # the contract might selfdestruct. on the other hand we can omit this
        # when we _do_ expect return data because we later check
        # `returndatasize` (that check works even if the contract
        # selfdestructs).
        ret.append(_extcodesize_check(contract_address))

    gas = call_kwargs.gas
    value = call_kwargs.value

    use_staticcall = fn_type.mutability in (StateMutability.VIEW, StateMutability.PURE)
    if context.is_constant():
        assert use_staticcall, "typechecker missed this"

    if use_staticcall:
        call_op = ["staticcall", gas, contract_address, args_ofst, args_len, buf, ret_len]
    else:
        call_op = ["call", gas, contract_address, value, args_ofst, args_len, buf, ret_len]

    ret.append(check_external_call(call_op))

    return_t = fn_type.return_type
    if return_t is not None:
        ret.append(ret_unpacker)

    return IRnode.from_list(ret, typ=return_t, location=MEMORY)


def ir_for_external_call(call_expr, context):
    from vyper.codegen.expr import Expr  # TODO rethink this circular import

    contract_address = Expr.parse_value_expr(call_expr.func.value, context)
    assert isinstance(contract_address.typ, InterfaceT)
    args_ir = [Expr(x, context).ir_node for x in call_expr.args]
    call_kwargs = _parse_kwargs(call_expr, context)

    with contract_address.cache_when_complex("external_contract") as (b1, contract_address):
        return b1.resolve(
            _external_call_helper(contract_address, args_ir, call_kwargs, call_expr, context)
        )



import pytest

from vyper import compiler
from vyper.exceptions import InvalidLiteral, StructureException, TypeMismatch

fail_list = [
    (
        """
@external
def foo():
    x: int128[3] = [1, 2, 3]
    x = 4
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo():
    x: int128[3] = [1, 2, 3]
    x = [4, 5, 6, 7]
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo() -> int128[2]:
    return [3, 5, 7]
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo() -> int128[2]:
    return [3]
    """,
        TypeMismatch,
    ),
    (
        """
y: int128[3]

@external
def foo(x: int128[3]):
    self.y = x[0]
    """,
        TypeMismatch,
    ),
    (
        """
y: int128[3]

@external
def foo(x: int128[3]):
    self.y[0] = x
    """,
        TypeMismatch,
    ),
    (
        """
y: int128[4]

@external
def foo(x: int128[3]):
    self.y = x
    """,
        TypeMismatch,
    ),
    (
        """
y: decimal[2]

@external
def foo(x: int128[2][2]):
    self.y = x[1]
    """,
        TypeMismatch,
    ),
    (
        """
bar: int128[3]
@external
def foo():
    self.bar = [1, 2, 0x1234567890123456789012345678901234567890]
    """,
        InvalidLiteral,
    ),
    (
        """
bar: int128[3]
@external
def foo():
    self.bar = []
    """,
        TypeMismatch,
    ),
    (
        """
b: int128[5]
@external
def foo():
    x: int128 = self.b[0][1]
    """,
        StructureException,
    ),
    (
        """
bar: int128[3]
@external
def foo():
    self.bar = [1, [2], 3]
    """,
        InvalidLiteral,
    ),
    (
        """
bar: int128[3][3]
@external
def foo():
    self.bar = 5
    """,
        TypeMismatch,
    ),
    (
        """
bar: int128[3][3]
@external
def foo():
    self.bar = [2, 5]
    """,
        TypeMismatch,
    ),
    (
        """
bar: int128[3]
@external
def foo():
    self.bar = [1, 2, 3, 4]
    """,
        TypeMismatch,
    ),
    (
        """
bar: int128[3]
@external
def foo():
    self.bar = [1, 2]
    """,
        TypeMismatch,
    ),
    (
        """
b: int128[5]
@external
def foo():
    self.b[0] = 7.5
    """,
        TypeMismatch,
    ),
    (
        """
b: int128[5]
@external
def foo():
    x: int128[5] = self.b[0].cow
    """,
        StructureException,
    ),
    (
        """
@external
def foo()->bool[2]:
    a: decimal[2] = [0.0, 0.0]
    a[0] = 1
    return a
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo()->bool[2]:
    a: bool[10] = [True, True, True, True, True, True, True, True, True, True]
    a[0] = 1
    return a
    """,
        TypeMismatch,
    ),
    (
        """
@external
def test() -> int128:
    a: int128[3] = [1, 2, 3.0]
    return a[0]
    """,
        InvalidLiteral,
    ),
    (
        """
@external
def test() -> int128:
    a: int128[3] = [1, 2, True]
    return a[0]
    """,
        InvalidLiteral,
    ),
    (
        """
y: decimal[3]

@external
def foo(x: int128[3]):
    self.y = x
    """,
        TypeMismatch,
    ),
    (
        """
y: decimal[2][2]

@external
def foo(x: int128[2][2]):
    self.y = x
    """,
        TypeMismatch,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_block_fail(bad_code, exc):
    with pytest.raises(exc):
        compiler.compile_code(bad_code)


valid_list = [
    """
@external
def foo():
    x: int128[3] = [1, 2, 3]
    x = [4, 5, 6]
    """,
    """
@external
def foo() -> int128[2][2]:
    return [[1,2], [3,4]]
    """,
    """
@external
def foo() -> decimal[2][2]:
    return [[1.0, 2.0], [3.0, 4.0]]
    """,
    """
@external
def foo() -> decimal[2][2]:
    return [[1.0, 2.0], [3.5, 4.0]]
    """,
    """
@external
def foo(x: int128[3]) -> int128:
    return x[0]
    """,
    """
y: int128[3]

@external
def foo(x: int128[3]):
    self.y = x
    """,
    """
@external
def foo() -> int128[2]:
    return [3,5]
    """,
    """
bar: decimal[3]
@external
def foo():
    self.bar = [1.0, 2.1, 3.0]
    """,
    """
x: int128[1][2][3][4][5]
    """,
    """
bar: int128[3]
@external
def foo():
    self.bar = [1, 2, 3]
    """,
    """
b: int128[5]
@external
def foo():
    a: int128[5] = [0, 0, 0, 0, 0]
    self.b[0] = a[0]
    """,
    """
b: decimal[5]
@external
def foo():
    self.b[0] = 7.0
    """,
    """
@external
def foo():
    x: DynArray[uint256, 3] = [1, 2, 3]
    for i: DynArray[uint256, 3] in [[], []]:
        x = i
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_list_success(good_code):
    assert compiler.compile_code(good_code) is not None



import pytest

from vyper import compiler
from vyper.exceptions import StructureException

fail_list = [
    """
@external
def foo() -> int128:
    x: int128 = 45
    return x.codesize
    """
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_block_fail(bad_code):
    with pytest.raises(StructureException):
        compiler.compile_code(bad_code)


valid_list = [
    """
@external
def foo() -> uint256:
    x: address = 0x1234567890123456789012345678901234567890
    return x.codesize
    """,
    """
@external
def foo() -> uint256:
    return self.codesize
    """,
    """
struct Foo:
    t: address
foo: Foo

@external
def bar() -> uint256:
    return self.foo.t.codesize
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_block_success(good_code):
    assert compiler.compile_code(good_code) is not None



import contextlib
from typing import Any

from vyper.exceptions import VyperException


# a commit/rollback scheme for metadata caching. in the case that an
# exception is thrown and caught during type checking (currently, only
# during for loop iterator variable type inference), we can roll back
# any state updates due to type checking.
# this is implemented as a stack of changesets, because we need to
# handle nested rollbacks in the case of nested for loops
class _NodeMetadataJournal:
    _NOT_FOUND = object()

    def __init__(self):
        self._node_updates: list[dict[tuple[int, str, Any], NodeMetadata]] = []

    def register_update(self, metadata, k):
        KEY = (id(metadata), k)
        if KEY in self._node_updates[-1]:
            return
        prev = metadata.get(k, self._NOT_FOUND)
        self._node_updates[-1][KEY] = (metadata, prev)

    @contextlib.contextmanager
    def enter(self):
        self._node_updates.append({})
        try:
            yield
        except VyperException as e:
            # note: would be better to only catch typechecker exceptions here.
            self._rollback_inner()
            raise e from e
        else:
            self._commit_inner()

    def _rollback_inner(self):
        for (_, k), (metadata, prev) in self._node_updates[-1].items():
            if prev is self._NOT_FOUND:
                metadata.pop(k, None)
            else:
                metadata[k] = prev
        self._pop_inner()

    def _commit_inner(self):
        inner = self._pop_inner()

        if len(self._node_updates) == 0:
            return

        outer = self._node_updates[-1]

        # register with previous frame in case inner gets committed
        # but outer needs to be rolled back
        for (_, k), (metadata, prev) in inner.items():
            if (id(metadata), k) not in outer:
                outer[(id(metadata), k)] = (metadata, prev)

    def _pop_inner(self):
        return self._node_updates.pop()


class NodeMetadata(dict):
    """
    A data structure which allows for journaling.
    """

    _JOURNAL: _NodeMetadataJournal = _NodeMetadataJournal()

    def __setitem__(self, k, v):
        # if we are in a context where we need to journal, add
        # this to the changeset.
        if len(self._JOURNAL._node_updates) != 0:
            self._JOURNAL.register_update(self, k)

        super().__setitem__(k, v)

    @classmethod
    @contextlib.contextmanager
    def enter_typechecker_speculation(cls):
        with cls._JOURNAL.enter():
            yield



import pytest

from tests.utils import ZERO_ADDRESS
from vyper.compiler import compile_code
from vyper.exceptions import (
    InvalidLiteral,
    NonPayableViolation,
    StateAccessViolation,
    TypeMismatch,
    UndeclaredDefinition,
)
from vyper.utils import method_id


def test_default_param_abi(get_contract):
    code = """
@external
@payable
def safeTransferFrom(_data: Bytes[100] = b"test", _b: int128 = 1):
    pass
    """
    abi = get_contract(code).abi

    assert len(abi) == 3
    assert set([fdef["name"] for fdef in abi]) == {"safeTransferFrom"}
    assert abi[0]["inputs"] == []
    assert abi[1]["inputs"] == [{"type": "bytes", "name": "_data"}]
    assert abi[2]["inputs"] == [
        {"type": "bytes", "name": "_data"},
        {"type": "int128", "name": "_b"},
    ]


def test_basic_default_param_passthrough(get_contract):
    code = """
@external
def fooBar(_data: Bytes[100] = b"test", _b: int128 = 1) -> int128:
    return 12321
    """

    c = get_contract(code)

    assert c.fooBar() == 12321
    assert c.fooBar(b"drum drum") == 12321
    assert c.fooBar(b"drum drum", 2) == 12321


def test_basic_default_param_set(get_contract):
    code = """
@external
def fooBar(a:int128, b: uint256 = 333) -> (int128, uint256):
    return a, b
    """

    c = get_contract(code)
    assert c.fooBar(456, 444) == (456, 444)
    assert c.fooBar(456) == (456, 333)


def test_basic_default_param_set_2args(get_contract):
    code = """
@external
def fooBar(a:int128, b: uint256 = 999, c: address = 0x0000000000000000000000000000000000000001) -> (int128, uint256, address):  # noqa: E501
    return a, b, c
    """

    c = get_contract(code)
    c_default_value = "0x0000000000000000000000000000000000000001"
    b_default_value = 999
    addr2 = "0x1000000000000000000000000000000000004321"

    # b default value, c default value
    assert c.fooBar(123) == (123, b_default_value, c_default_value)
    # c default_value, b set from param
    assert c.fooBar(456, 444) == (456, 444, c_default_value)
    # no default values
    assert c.fooBar(6789, 4567, addr2) == (6789, 4567, addr2)


def test_default_param_bytes(get_contract):
    code = """
@external
def fooBar(a: Bytes[100], b: int128, c: Bytes[100] = b"testing", d: uint256 = 999) -> (Bytes[100], int128, Bytes[100], uint256):  # noqa: E501
    return a, b, c, d
    """
    c = get_contract(code)
    c_default = b"testing"
    d_default = 999

    # c set, 7d default value
    assert c.fooBar(b"booo", 12321, b"woo") == (b"booo", 12321, b"woo", d_default)
    # d set, c set
    assert c.fooBar(b"booo", 12321, b"lucky", 777) == (b"booo", 12321, b"lucky", 777)
    # no default values
    assert c.fooBar(b"booo", 12321) == (b"booo", 12321, c_default, d_default)


def test_default_param_array(get_contract):
    code = """
@external
def fooBar(a: Bytes[100], b: uint256[2], c: Bytes[6] = b"hello", d: int128[3] = [6, 7, 8]) -> (Bytes[100], uint256, Bytes[6], int128):  # noqa: E501
    return a, b[1], c, d[2]
    """
    c = get_contract(code)
    c_default = b"hello"
    d_default = 8

    # c set, d default value
    assert c.fooBar(b"booo", [99, 88], b"woo") == (b"booo", 88, b"woo", d_default)
    # d set, c set
    assert c.fooBar(b"booo", [22, 11], b"lucky", [24, 25, 26]) == (b"booo", 11, b"lucky", 26)
    # no default values
    assert c.fooBar(b"booo", [55, 66]) == (b"booo", 66, c_default, d_default)


def test_default_param_interface(get_contract):
    code = """
interface Foo:
    def bar(): payable

FOO: constant(Foo) = Foo(0xFFfFfFffFFfffFFfFFfFFFFFffFFFffffFfFFFfF)

@external
def bar(a: uint256, b: Foo = Foo(0xF5D4020dCA6a62bB1efFcC9212AAF3c9819E30D7)) -> Foo:
    return b

@external
def baz(a: uint256, b: Foo = Foo(empty(address))) -> Foo:
    return b

@external
def faz(a: uint256, b: Foo = FOO) -> Foo:
    return b
    """
    c = get_contract(code)

    addr1 = "0xFFfFfFffFFfffFFfFFfFFFFFffFFFffffFfFFFfF"
    addr2 = "0xF5D4020dCA6a62bB1efFcC9212AAF3c9819E30D7"

    assert c.bar(1) == addr2
    assert c.bar(1, addr1) == addr1
    assert c.baz(1) == ZERO_ADDRESS
    assert c.baz(1, ZERO_ADDRESS) == ZERO_ADDRESS
    assert c.faz(1) == addr1
    assert c.faz(1, addr1) == addr1


def test_default_param_internal_function(get_contract):
    code = """
@internal
@view
def _foo(a: int128[3] = [1, 2, 3]) -> int128[3]:
    b: int128[3] = a
    return b


@external
@view
def foo() -> int128[3]:
    return self._foo([4, 5, 6])

@external
@view
def foo2() -> int128[3]:
    return self._foo()
    """
    c = get_contract(code)

    assert c.foo() == [4, 5, 6]
    assert c.foo2() == [1, 2, 3]


def test_default_param_external_function(get_contract):
    code = """
@external
@view
def foo(a: int128[3] = [1, 2, 3]) -> int128[3]:
    b: int128[3] = a
    return b
    """
    c = get_contract(code)

    assert c.foo([4, 5, 6]) == [4, 5, 6]
    assert c.foo() == [1, 2, 3]


def test_default_param_clamp(env, get_contract, tx_failed):
    code = """
@external
def bar(a: int128, b: int128 = -1) -> (int128, int128):  # noqa: E501
    return a, b
    """

    c = get_contract(code)

    assert c.bar(-123) == (-123, -1)
    assert c.bar(100, 100) == (100, 100)

    method = method_id("bar(int128,int128)")
    good_data = (200).to_bytes(32, "big") + (2**127 - 1).to_bytes(32, "big")
    bad_data = (200).to_bytes(32, "big") + (2**127).to_bytes(32, "big")

    assert env.message_call(c.address, data=method + good_data) == good_data
    with tx_failed():
        env.message_call(c.address, data=method + bad_data)


def test_default_param_private(get_contract):
    code = """
@internal
def fooBar(a: Bytes[100], b: uint256, c: Bytes[20] = b"crazy") -> (Bytes[100], uint256, Bytes[20]):
    return a, b, c

@external
def callMe() -> (Bytes[100], uint256, Bytes[20]):
    return self.fooBar(b'I just met you', 123456)

@external
def callMeMaybe() -> (Bytes[100], uint256, Bytes[20]):
    # return self.fooBar(b'here is my number', 555123456, b'baby')
    a: Bytes[100] = b""
    b: uint256 = 0
    c: Bytes[20] = b""
    a, b, c = self.fooBar(b'here is my number', 555123456, b'baby')
    return a, b, c
    """

    c = get_contract(code)

    # assert c.callMe() == (b'hello there', 123456, b'crazy')
    assert c.callMeMaybe() == (b"here is my number", 555123456, b"baby")


def test_environment_vars_as_default(get_contract, env):
    code = """
xx: uint256

@external
@payable
def foo(a: uint256 = msg.value) -> bool:
    self.xx += a
    return True

@external
def bar() -> uint256:
    return self.xx

@external
def get_balance() -> uint256:
    return self.balance
    """
    c = get_contract(code)
    env.set_balance(env.deployer, 31337 + 9001)
    c.foo(value=31337)
    assert c.bar() == 31337
    c.foo(666, value=9001)
    assert c.bar() == 31337 + 666
    assert c.get_balance() == 31337 + 9001
    assert env.get_balance(env.deployer) == 0


PASSING_CONTRACTS = [
    """
@external
def foo(a: bool = True, b: bool[2] = [True, False]): pass
    """,
    """
@external
def foo(
    a: address = 0x0c04792e92e6b2896a18568fD936781E9857feB7,
    b: address[2] = [
        0x0c04792e92e6b2896a18568fD936781E9857feB7,
        0x0c04792e92e6b2896a18568fD936781E9857feB7
    ]): pass
    """,
    """
@external
def foo(a: uint256 = 12345, b: uint256[2] = [31337, 42]): pass
    """,
    """
@external
def foo(a: int128 = -31, b: int128[2] = [64, -46]): pass
    """,
    """
@external
def foo(a: Bytes[6] = b"potato"): pass
    """,
    """
@external
def foo(a: decimal = 3.14, b: decimal[2] = [1.337, 2.69]): pass
    """,
    """
@external
def foo(a: address = msg.sender, b: address[3] = [msg.sender, tx.origin, block.coinbase]): pass
    """,
    """
@internal
def foo(a: address = msg.sender, b: address[3] = [msg.sender, tx.origin, block.coinbase]): pass
    """,
    """
@external
@payable
def foo(a: uint256 = msg.value): pass
    """,
    """
@external
def foo(a: uint256 = 2**8): pass
    """,
    """
struct Bar:
    a: address
    b: uint256

@external
def foo(bar: Bar = Bar(a=msg.sender, b=1)): pass
    """,
    """
struct Baz:
    c: address
    d: int128

struct Bar:
    a: address
    b: Baz

@external
def foo(bar: Bar = Bar(a=msg.sender, b=Baz(c=block.coinbase, d=-10))): pass
    """,
    """
A: public(address)

@external
def foo(a: address = empty(address)):
    self.A = a
    """,
    """
A: public(int112)

@external
def foo(a: int112 = min_value(int112)):
    self.A = a
    """,
    """
struct X:
    x: int128
    y: address
BAR: constant(X) = X(x=1, y=0x0000000000000000000000000000000000012345)
@external
def out_literals(a: int128 = BAR.x + 1) -> X:
    return BAR
    """,
    """
struct X:
    x: int128
    y: address
struct Y:
    x: X
    y: uint256
BAR: constant(X) = X(x=1, y=0x0000000000000000000000000000000000012345)
FOO: constant(Y) = Y(x=BAR, y=256)
@external
def out_literals(a: int128 = FOO.x.x + 1) -> Y:
    return FOO
    """,
    """
struct Bar:
    a: bool

BAR: constant(Bar) = Bar(a=True)

@external
def foo(x: bool = True and not BAR.a):
    pass
    """,
    """
struct Bar:
    a: uint256

BAR: constant(Bar) = Bar(a=123)

@external
def foo(x: bool = BAR.a + 1 > 456):
    pass
    """,
]


@pytest.mark.parametrize("code", PASSING_CONTRACTS)
def test_good_default_params(code):
    assert compile_code(code)


FAILING_CONTRACTS = [
    (
        """
# default params must be literals
x: int128

@external
def foo(xx: int128, y: int128 = xx): pass
    """,
        UndeclaredDefinition,
    ),
    (
        """
# value out of range for uint256
@external
def foo(a: uint256 = -1): pass
    """,
        TypeMismatch,
    ),
    (
        """
# value out of range for int128
@external
def foo(a: int128 = 170141183460469231731687303715884105728): pass
    """,
        TypeMismatch,
    ),
    (
        """
# value out of range for uint256 array
@external
def foo(a: uint256[2] = [13, -42]): pass
     """,
        TypeMismatch,
    ),
    (
        """
# value out of range for int128 array
@external
def foo(a: int128[2] = [12, 170141183460469231731687303715884105728]): pass
    """,
        TypeMismatch,
    ),
    (
        """
# array type mismatch
@external
def foo(a: uint256[2] = [12, True]): pass
    """,
        InvalidLiteral,
    ),
    (
        """
# wrong length
@external
def foo(a: uint256[2] = [1, 2, 3]): pass
    """,
        TypeMismatch,
    ),
    (
        """
# default params must be literals
x: uint256

@external
def foo(a: uint256 = self.x): pass
     """,
        StateAccessViolation,
    ),
    (
        """
# default params must be literals inside array
x: uint256

@external
def foo(a: uint256[2] = [2, self.x]): pass
     """,
        StateAccessViolation,
    ),
    (
        """
# msg.value in a nonpayable
@external
def foo(a: uint256 = msg.value): pass
""",
        NonPayableViolation,
    ),
]


@pytest.mark.parametrize("failing_contract", FAILING_CONTRACTS)
def test_bad_default_params(failing_contract, assert_compile_failed):
    code, exc = failing_contract
    assert_compile_failed(lambda: compile_code(code), exc)



import pytest
from pytest import raises

from vyper import compiler
from vyper.exceptions import (
    InvalidAttribute,
    TypeMismatch,
    UndeclaredDefinition,
    UnknownAttribute,
    VariableDeclarationException,
)

fail_list = [
    (
        """
@external
def test():
    a = 1
    """,
        UndeclaredDefinition,
    ),
    (
        """
@external
def test():
    a = 33.33
    """,
        UndeclaredDefinition,
    ),
    (
        """
@external
def test():
    a = "test string"
    """,
        UndeclaredDefinition,
    ),
    (
        """
@external
def test():
    a: int128 = 33.33
    """,
        TypeMismatch,
    ),
    (
        """
@external
def data() -> int128:
    s: int128[5] = [1, 2, 3, 4, 5, 6]
    return 235357
    """,
        TypeMismatch,
    ),
    (
        """
struct S:
    a: int128
    b: decimal
@external
def foo() -> int128:
    s: S = S(a=1.2, b=1)
    return s.a
    """,
        TypeMismatch,
    ),
    (
        """
struct S:
    a: int128
    b: decimal
@external
def foo() -> int128:
    s: S = S(a=1)
    """,
        VariableDeclarationException,
    ),
    (
        """
struct S:
    a: int128
    b: decimal
@external
def foo() -> int128:
    s: S = S(b=1.2, a=1)
    """,
        InvalidAttribute,
    ),
    (
        """
struct S:
    a: int128
    b: decimal
@external
def foo() -> int128:
    s: S = S(a=1, b=1.2, c=1, d=33, e=55)
    return s.a
    """,
        UnknownAttribute,
    ),
    (
        """
@external
def foo() -> bool:
    a: uint256 = -1
    return True
""",
        TypeMismatch,
    ),
    (
        """
@external
def foo() -> bool:
    a: uint256[2] = [13, -42]
    return True
""",
        TypeMismatch,
    ),
    (
        """
@external
def foo() -> bool:
    a: int128 = 170141183460469231731687303715884105728
    return True
""",
        TypeMismatch,
    ),
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_as_wei_fail(bad_code):
    with raises(bad_code[1]):
        compiler.compile_code(bad_code[0])


valid_list = [
    """
struct S:
    a: int128
    b: decimal
@internal
def do_stuff() -> bool:
    return True

@external
def test():
    a: bool = self.do_stuff() or self.do_stuff()
    """,
    """
@internal
def do_stuff() -> bool:
    return True

@external
def test():
    a: bool = False or self.do_stuff()
    """,
    """
@external
def test():
    a: int128 = 1
    """,
    """
@internal
def do_stuff() -> bool:
    return True

@external
def test():
    a: bool = self.do_stuff()
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_ann_assign_success(good_code):
    assert compiler.compile_code(good_code) is not None



import pytest

from vyper.exceptions import TypeMismatch


def test_test_bytes(get_contract, tx_failed):
    test_bytes = """
@external
def foo(x: Bytes[100]) -> Bytes[100]:
    return x
    """

    c = get_contract(test_bytes)
    moo_result = c.foo(b"cow")
    assert moo_result == b"cow"

    print("Passed basic bytes test")

    assert c.foo(b"\x35" * 100) == b"\x35" * 100

    print("Passed max-length bytes test")

    # test for greater than 100 bytes, should raise exception
    with tx_failed():
        c.foo(b"\x35" * 101)

    print("Passed input-too-long test")


def test_test_bytes2(get_contract):
    test_bytes2 = """
@external
def foo(x: Bytes[100]) -> Bytes[100]:
    y: Bytes[100] = x
    return y
    """

    c = get_contract(test_bytes2)
    assert c.foo(b"cow") == b"cow"
    assert c.foo(b"") == b""
    assert c.foo(b"\x35" * 63) == b"\x35" * 63
    assert c.foo(b"\x35" * 64) == b"\x35" * 64
    assert c.foo(b"\x35" * 65) == b"\x35" * 65

    print("Passed string copying test")


def test_test_bytes3(get_contract):
    test_bytes3 = """
x: int128
maa: Bytes[60]
y: int128

@deploy
def __init__():
    self.x = 27
    self.y = 37

@external
def set_maa(inp: Bytes[60]):
    self.maa = inp

@external
def set_maa2(inp: Bytes[60]):
    ay: Bytes[60] = inp
    self.maa = ay

@external
def get_maa() -> Bytes[60]:
    return self.maa

@external
def get_maa2() -> Bytes[60]:
    ay: Bytes[60] = self.maa
    return ay

@external
def get_xy() -> int128:
    return self.x * self.y
    """

    c = get_contract(test_bytes3)
    c.set_maa(b"pig")
    assert c.get_maa() == b"pig"
    assert c.get_maa2() == b"pig"
    c.set_maa2(b"")
    assert c.get_maa() == b""
    assert c.get_maa2() == b""
    c.set_maa(b"\x44" * 60)
    assert c.get_maa() == b"\x44" * 60
    assert c.get_maa2() == b"\x44" * 60
    c.set_maa2(b"mongoose")
    assert c.get_maa() == b"mongoose"
    assert c.get_xy() == 999

    print("Passed advanced string copying test")


def test_test_bytes4(get_contract):
    test_bytes4 = """
a: Bytes[60]
@external
def foo(inp: Bytes[60]) -> Bytes[60]:
    self.a = inp
    self.a = b""
    return self.a

@external
def bar(inp: Bytes[60]) -> Bytes[60]:
    b: Bytes[60] = inp
    b = b""
    return b
    """

    c = get_contract(test_bytes4)
    assert c.foo(b"") == b"", c.foo()
    assert c.bar(b"") == b""

    print("Passed string deleting test")


def test_test_bytes5(get_contract):
    test_bytes5 = """
struct G:
    a: Bytes[50]
    b: Bytes[50]
struct H:
    a: Bytes[40]
    b: Bytes[45]

g: G

@external
def foo(inp1: Bytes[40], inp2: Bytes[45]):
    self.g = G(a=inp1, b=inp2)

@external
def check1() -> Bytes[50]:
    return self.g.a

@external
def check2() -> Bytes[50]:
    return self.g.b

@external
def bar(inp1: Bytes[40], inp2: Bytes[45]) -> Bytes[50]:
    h: H = H(a=inp1, b=inp2)
    return h.a

@external
def bat(inp1: Bytes[40], inp2: Bytes[45]) -> Bytes[50]:
    h: H = H(a=inp1, b=inp2)
    return h.b

@external
def quz(inp1: Bytes[40], inp2: Bytes[45]):
    h:  H = H(a=inp1, b=inp2)
    self.g.a = h.a
    self.g.b = h.b
    """

    c = get_contract(test_bytes5)
    c.foo(b"cow", b"horse")
    assert c.check1() == b"cow"
    assert c.check2() == b"horse"
    assert c.bar(b"pig", b"moose") == b"pig"
    assert c.bat(b"pig", b"moose") == b"moose"
    c.quz(b"badminton", b"fluffysheep")
    assert c.check1() == b"badminton"
    assert c.check2() == b"fluffysheep"

    print("Passed string struct test")


def test_binary_literal(get_contract):
    bytes_to_num_code = """
r: Bytes[1]

@external
def get(a: Bytes[1]) -> Bytes[2]:
    return concat(a, 0b00000001)

@external
def getsome() -> Bytes[1]:
    return 0b00001110

@external
def testsome(a: Bytes[1]) -> bool:
    return a == 0b01100001

@external
def testsome_storage(y: Bytes[1]) -> bool:
    self.r = 0b01100001
    return self.r == y
    """

    c = get_contract(bytes_to_num_code)

    assert c.getsome() == b"\x0e"
    assert c.testsome(b"a")
    assert c.testsome(b"\x61")
    assert c.testsome(0b1100001.to_bytes(1, "big"))
    assert not c.testsome(b"b")
    assert c.testsome_storage(b"a")
    assert not c.testsome_storage(b"x")


def test_bytes_comparison(get_contract):
    code = """
@external
def get_mismatch(a: Bytes[1]) -> bool:
    b: Bytes[2] = b'ab'
    return a == b

@external
def get_large(a: Bytes[100]) -> bool:
    b: Bytes[100] = b'ab'
    return a == b
    """

    c = get_contract(code)
    assert c.get_mismatch(b"\x00") is False
    assert c.get_large(b"\x00") is False
    assert c.get_large(b"ab") is True


def test_bytes32_literals(get_contract):
    code = """
@external
def test() -> bool:
    l: bytes32 = 0x0000000000000000000000000000000000000000000000000000000080ac58cd
    return l == 0x0000000000000000000000000000000000000000000000000000000080ac58cd

    """

    c = get_contract(code)

    assert c.test() is True


@pytest.mark.parametrize("m,val", [(2, b"ab"), (3, b"ab"), (3, b"abc")])
def test_bytes_literals(get_contract, m, val):
    vyper_literal = "0x" + val.ljust(m, b"\x00").hex()
    code = f"""
@external
def test() -> bool:
    l: bytes{m} = {vyper_literal}
    return l == {vyper_literal}

@external
def test2(l: bytes{m} = {vyper_literal}) -> bool:
    return l == {vyper_literal}
    """

    c = get_contract(code)

    assert c.test() is True
    assert c.test2() is True
    assert c.test2(vyper_literal) is True


def test_zero_padding_with_private(get_contract):
    code = """
counter: uint256

@internal
@view
def to_little_endian_64(_value: uint256) -> Bytes[8]:
    y: uint256 = 0
    x: uint256 = _value
    for _: uint256 in range(8):
        y = (y << 8) | (x & 255)
        x >>= 8
    return slice(convert(y, bytes32), 24, 8)

@external
def set_count(i: uint256):
    self.counter = i

@external
@view
def get_count() -> Bytes[24]:
    return self.to_little_endian_64(self.counter)
    """

    c = get_contract(code)

    assert c.get_count() == b"\x00\x00\x00\x00\x00\x00\x00\x00"
    c.set_count(1)
    assert c.get_count() == b"\x01\x00\x00\x00\x00\x00\x00\x00"
    c.set_count(0xF0F0F0)
    assert c.get_count() == b"\xf0\xf0\xf0\x00\x00\x00\x00\x00"
    c.set_count(0x0101010101010101)
    assert c.get_count() == b"\x01\x01\x01\x01\x01\x01\x01\x01"


cases_invalid_assignments = [
    (
        """
@external
def assign():
    xs: Bytes[32] = b"abcdef"
    y: bytes32 = xs
    """,
        TypeMismatch,
    ),
    (
        """
@external
def assign():
    xs: bytes6 = b"abcdef"
    """,
        TypeMismatch,
    ),
    (
        """
@external
def assign():
    xs: bytes4 = 0xabcdef  # bytes3 literal
    """,
        TypeMismatch,
    ),
    (
        """
@external
def assign():
    xs: bytes4 = 0x1234abcdef # bytes5 literal
    """,
        TypeMismatch,
    ),
]


@pytest.mark.parametrize("code,exc", cases_invalid_assignments)
def test_invalid_assignments(get_contract, assert_compile_failed, code, exc):
    assert_compile_failed(lambda: get_contract(code), exc)



import pytest

from vyper import compile_code
from vyper.exceptions import OverflowException, TypeMismatch


def test_state_accessor(get_contract):
    state_accessor = """
y: HashMap[int128, int128]

@external
def oo():
    self.y[3] = 5

@external
def foo() -> int128:
    return self.y[3]

    """

    c = get_contract(state_accessor)
    c.oo()
    assert c.foo() == 5


def test_getter_code(get_contract):
    getter_code = """
interface V:
    def foo(): nonpayable

struct W:
    a: uint256
    b: int128[7]
    c: Bytes[100]
    e: int128[3][3]
    f: uint256
    g: uint256
x: public(uint256)
y: public(int128[5])
z: public(Bytes[100])
w: public(HashMap[int128, W])
a: public(uint256[10][10])
b: public(HashMap[uint256, HashMap[address, uint256[4]]])
c: public(constant(uint256)) = 1
d: public(immutable(uint256))
e: public(immutable(uint256[2]))
f: public(constant(uint256[2])) = [3, 7]
g: public(constant(V)) = V(0xFFfFfFffFFfffFFfFFfFFFFFffFFFffffFfFFFfF)

@deploy
def __init__():
    self.x = as_wei_value(7, "wei")
    self.y[1] = 9
    self.z = b"cow"
    self.w[1].a = 11
    self.w[1].b[2] = 13
    self.w[1].c = b"horse"
    self.w[2].e[1][2] = 17
    self.w[3].f = 750
    self.w[3].g = 751
    self.a[1][4] = 666
    self.b[42][self] = [5,6,7,8]
    d = 1729
    e = [2, 3]
    """

    c = get_contract(getter_code)
    assert c.x() == 7
    assert c.y(1) == 9
    assert c.z() == b"cow"
    assert c.w(1)[0] == 11  # W.a
    assert c.w(1)[1][2] == 13  # W.b[2]
    assert c.w(1)[2] == b"horse"  # W.c
    assert c.w(2)[3][1][2] == 17  # W.e[1][2]
    assert c.w(3)[4] == 750  # W.f
    assert c.w(3)[5] == 751  # W.g
    assert c.a(1, 4) == 666
    assert c.b(42, c.address, 2) == 7
    assert c.c() == 1
    assert c.d() == 1729
    assert c.e(0) == 2
    assert [c.f(i) for i in range(2)] == [3, 7]
    assert c.g() == "0xFFfFfFffFFfffFFfFFfFFFFFffFFFffffFfFFFfF"


def test_getter_mutability(get_contract):
    code = """
foo: public(uint256)
goo: public(String[69])
bar: public(uint256[4][5])
baz: public(HashMap[address, Bytes[100]])
potatoes: public(HashMap[uint256, HashMap[bytes32, uint256[4]]])
nyoro: public(constant(uint256)) = 2
kune: public(immutable(uint256))

@deploy
def __init__():
    kune = 2
"""

    contract = get_contract(code)

    for item in contract.abi:
        if item["type"] == "constructor":
            continue
        assert item["stateMutability"] == "view"


@pytest.mark.parametrize(
    "typ,index,expected_error",
    [
        ("uint256", "-1", TypeMismatch),
        ("uint256", "0-1", TypeMismatch),
        ("uint256", "0-1+1", TypeMismatch),
        ("uint256", "2**256", OverflowException),
        ("uint256", "2**256 // 2", OverflowException),
        ("uint256", "2 * 2**255", OverflowException),
        ("int256", "-2**255", TypeMismatch),
        ("int256", "-2**256", OverflowException),
        ("int256", "2**255", TypeMismatch),
        ("int256", "2**256 - 5", OverflowException),
        ("int256", "2 * 2**254", TypeMismatch),
        ("int8", "*".join(["2"] * 7), TypeMismatch),
    ],
)
def test_hashmap_index_checks(typ, index, expected_error):
    code = f"""
m: HashMap[{typ}, uint256]

@external
def foo():
    self.m[{index}] = 2
    """
    with pytest.raises(expected_error):
        compile_code(code)



import itertools
import operator
import random

import pytest

from vyper import compile_code
from vyper.exceptions import (
    InvalidOperation,
    OverflowException,
    TypeMismatch,
    ZeroDivisionException,
)
from vyper.semantics.types import IntegerT
from vyper.utils import SizeLimits, evm_div, evm_mod

types = sorted(IntegerT.unsigneds())


@pytest.mark.parametrize("typ", types)
def test_exponent_base_zero(get_contract, typ):
    code = f"""
@external
def foo(x: {typ}) -> {typ}:
    return 0 ** x
    """
    lo, hi = typ.ast_bounds
    c = get_contract(code)
    assert c.foo(0) == 1
    assert c.foo(1) == 0
    assert c.foo(42) == 0
    assert c.foo(hi) == 0


@pytest.mark.parametrize("typ", types)
def test_exponent_base_one(get_contract, typ):
    code = f"""
@external
def foo(x: {typ}) -> {typ}:
    return 1 ** x
    """
    lo, hi = typ.ast_bounds
    c = get_contract(code)
    assert c.foo(0) == 1
    assert c.foo(1) == 1
    assert c.foo(42) == 1
    assert c.foo(hi) == 1


@pytest.mark.parametrize("typ", types)
def test_exponent_power_zero(get_contract, typ):
    # #2984
    code = f"""
@external
def foo(x: {typ}) -> {typ}:
    return x ** 0
    """
    lo, hi = typ.ast_bounds
    c = get_contract(code)
    assert c.foo(0) == 1
    assert c.foo(1) == 1
    assert c.foo(42) == 1
    assert c.foo(hi) == 1


@pytest.mark.parametrize("typ", types)
def test_exponent_power_one(get_contract, typ):
    # #2984
    code = f"""
@external
def foo(x: {typ}) -> {typ}:
    return x ** 1
    """
    lo, hi = typ.ast_bounds
    c = get_contract(code)
    assert c.foo(0) == 0
    assert c.foo(1) == 1
    assert c.foo(42) == 42
    assert c.foo(hi) == hi


ARITHMETIC_OPS = {
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "//": evm_div,
    "%": evm_mod,
}


@pytest.mark.parametrize("op", sorted(ARITHMETIC_OPS.keys()))
@pytest.mark.parametrize("typ", types)
@pytest.mark.fuzzing
def test_arithmetic_thorough(get_contract, tx_failed, op, typ):
    # both variables
    code_1 = f"""
@external
def foo(x: {typ}, y: {typ}) -> {typ}:
    return x {op} y
    """
    # right is literal
    code_2_template = """
@external
def foo(x: {typ}) -> {typ}:
    return x {op} {y}
    """
    # left is literal
    code_3_template = """
@external
def foo(y: {typ}) -> {typ}:
    return {x} {op} y
    """
    # both literals
    code_4_template = """
@external
def foo() -> {typ}:
    return {x} {op} {y}
    """

    fn = ARITHMETIC_OPS[op]
    c = get_contract(code_1)

    lo, hi = typ.ast_bounds
    bits = typ.bits

    special_cases = [0, 1, 2, 3, hi // 2 - 1, hi // 2, hi // 2 + 1, hi - 2, hi - 1, hi]
    xs = special_cases.copy()
    ys = special_cases.copy()
    NUM_CASES = 5
    # poor man's fuzzing - hypothesis doesn't make it easy
    # with the parametrized strategy
    xs += [random.randrange(lo, hi) for _ in range(NUM_CASES)]
    ys += [random.randrange(lo, hi) for _ in range(NUM_CASES)]

    # mirror signed integer tests
    assert 2 ** (bits - 1) in xs and (2**bits) - 1 in ys

    for x, y in itertools.product(xs, ys):
        expected = fn(x, y)

        in_bounds = lo <= expected <= hi
        # safediv and safemod disallow divisor == 0
        div_by_zero = y == 0 and op in ("//", "%")

        ok = in_bounds and not div_by_zero

        code_2 = code_2_template.format(typ=typ, op=op, y=y)
        code_3 = code_3_template.format(typ=typ, op=op, x=x)
        code_4 = code_4_template.format(typ=typ, op=op, x=x, y=y)

        if ok:
            assert c.foo(x, y) == expected
            assert get_contract(code_2).foo(x) == expected
            assert get_contract(code_3).foo(y) == expected
            assert get_contract(code_4).foo() == expected
        elif div_by_zero:
            with tx_failed():
                c.foo(x, y)
            with pytest.raises(ZeroDivisionException):
                get_contract(code_2)
            with tx_failed():
                get_contract(code_3).foo(y)
            with pytest.raises(ZeroDivisionException):
                get_contract(code_4)
        else:
            with tx_failed():
                c.foo(x, y)
            with tx_failed():
                get_contract(code_2).foo(x)
            with tx_failed():
                get_contract(code_3).foo(y)
            with pytest.raises((TypeMismatch, OverflowException)):
                get_contract(code_4)


COMPARISON_OPS = {
    "==": operator.eq,
    "!=": operator.ne,
    ">": operator.gt,
    ">=": operator.ge,
    "<": operator.lt,
    "<=": operator.le,
}


@pytest.mark.parametrize("op", sorted(COMPARISON_OPS.keys()))
@pytest.mark.parametrize("typ", types)
@pytest.mark.fuzzing
def test_comparators(get_contract, op, typ):
    code_1 = f"""
@external
def foo(x: {typ}, y: {typ}) -> bool:
    return x {op} y
    """

    fn = COMPARISON_OPS[op]
    c = get_contract(code_1)

    lo, hi = typ.ast_bounds

    # note: folding is tested in tests/unit/ast/nodes

    special_cases = [0, 1, 2, 3, hi // 2 - 1, hi // 2, hi // 2 + 1, hi - 2, hi - 1, hi]
    xs = special_cases.copy()
    ys = special_cases.copy()

    for x, y in itertools.product(xs, ys):
        expected = fn(x, y)
        assert c.foo(x, y) is expected


@pytest.mark.parametrize("typ", types)
def test_uint_literal(get_contract, typ):
    lo, hi = typ.ast_bounds

    good_cases = [0, 1, 2, 3, hi // 2 - 1, hi // 2, hi // 2 + 1, hi - 1, hi]
    bad_cases = [-1, -2, -3, -hi // 2, -hi + 1, -hi]
    code_template = """
@external
def test() -> {typ}:
    o: {typ} = {val}
    return o
    """

    for val in good_cases:
        c = get_contract(code_template.format(typ=typ, val=val))
        assert c.test() == val

    for val in bad_cases:
        exc = (
            TypeMismatch
            if SizeLimits.MIN_INT256 <= val <= SizeLimits.MAX_UINT256
            else OverflowException
        )
        with pytest.raises(exc):
            compile_code(code_template.format(typ=typ, val=val))


@pytest.mark.parametrize("typ", types)
@pytest.mark.parametrize("op", ["/"])
def test_invalid_ops(get_contract, assert_compile_failed, typ, op):
    code = f"""
@external
def foo(x: {typ}, y: {typ}) -> {typ}:
    return x {op} y
    """
    assert_compile_failed(lambda: get_contract(code), InvalidOperation)


@pytest.mark.parametrize("typ", types)
@pytest.mark.parametrize("op", ["not", "-"])
def test_invalid_unary_ops(get_contract, assert_compile_failed, typ, op):
    code = f"""
@external
def foo(a: {typ}) -> {typ}:
    return {op} a
    """
    with pytest.raises(InvalidOperation):
        compile_code(code)


def test_binop_nested_intermediate_overflow():
    code = """
@external
def foo():
    a: uint256 = 2**255 * 2 // 10
    """
    with pytest.raises(OverflowException):
        compile_code(code)


def test_invalid_div():
    code = """
@external
def foo():
    a: uint256 = 5 / 9
    """
    with pytest.raises(InvalidOperation) as e:
        compile_code(code)

    assert e.value._hint == "did you mean `5 // 9`?"



from vyper.utils import OrderedSet
from vyper.venom.analysis.cfg import CFGAnalysis
from vyper.venom.analysis.dfg import DFGAnalysis
from vyper.venom.analysis.liveness import LivenessAnalysis
from vyper.venom.basicblock import IRBasicBlock, IRInstruction, IRVariable
from vyper.venom.function import IRFunction
from vyper.venom.passes.base_pass import IRPass


class Mem2Var(IRPass):
    """
    This pass promoted memory operations to variable operations, when possible.
    It does yet do any memory aliasing analysis, so it is conservative.
    """

    function: IRFunction
    defs: dict[IRVariable, OrderedSet[IRBasicBlock]]

    def run_pass(self):
        self.analyses_cache.request_analysis(CFGAnalysis)
        dfg = self.analyses_cache.request_analysis(DFGAnalysis)
        self.analyses_cache.request_analysis(LivenessAnalysis)

        self.var_name_count = 0
        for var, inst in dfg.outputs.items():
            if inst.opcode != "alloca":
                continue
            self._process_alloca_var(dfg, var)

        self.analyses_cache.invalidate_analysis(DFGAnalysis)
        self.analyses_cache.invalidate_analysis(LivenessAnalysis)

    def _process_alloca_var(self, dfg: DFGAnalysis, var: IRVariable):
        """
        Process alloca allocated variable. If it is only used by mstore/mload/return
        instructions, it is promoted to a stack variable. Otherwise, it is left as is.
        """
        uses = dfg.get_uses(var)
        if all([inst.opcode == "mload" for inst in uses]):
            return
        elif all([inst.opcode == "mstore" for inst in uses]):
            return
        elif all([inst.opcode in ["mstore", "mload", "return"] for inst in uses]):
            var_name = f"addr{var.name}_{self.var_name_count}"
            self.var_name_count += 1
            for inst in uses:
                if inst.opcode == "mstore":
                    inst.opcode = "store"
                    inst.output = IRVariable(var_name)
                    inst.operands = [inst.operands[0]]
                elif inst.opcode == "mload":
                    inst.opcode = "store"
                    inst.operands = [IRVariable(var_name)]
                elif inst.opcode == "return":
                    bb = inst.parent
                    idx = bb.instructions.index(inst)
                    bb.insert_instruction(
                        IRInstruction("mstore", [IRVariable(var_name), inst.operands[1]]), idx
                    )



from vyper.ast.metadata import NodeMetadata
from vyper.exceptions import VyperException


def test_metadata_journal_basic():
    m = NodeMetadata()

    m["x"] = 1
    assert m["x"] == 1


def test_metadata_journal_commit():
    m = NodeMetadata()

    with m.enter_typechecker_speculation():
        m["x"] = 1

    assert m["x"] == 1


def test_metadata_journal_exception():
    m = NodeMetadata()

    m["x"] = 1
    try:
        with m.enter_typechecker_speculation():
            m["x"] = 2
            m["x"] = 3

            assert m["x"] == 3
            raise VyperException("dummy exception")

    except VyperException:
        pass

    # rollback upon exception
    assert m["x"] == 1


def test_metadata_journal_rollback_inner():
    m = NodeMetadata()

    m["x"] = 1
    with m.enter_typechecker_speculation():
        m["x"] = 2

        try:
            with m.enter_typechecker_speculation():
                m["x"] = 3
                m["x"] = 4  # test multiple writes

                assert m["x"] == 4
                raise VyperException("dummy exception")

        except VyperException:
            pass

    assert m["x"] == 2


def test_metadata_journal_rollback_outer():
    m = NodeMetadata()

    m["x"] = 1
    try:
        with m.enter_typechecker_speculation():
            m["x"] = 2

            with m.enter_typechecker_speculation():
                m["x"] = 3
                m["x"] = 4  # test multiple writes

            assert m["x"] == 4

            m["x"] = 5

            raise VyperException("dummy exception")

    except VyperException:
        pass

    assert m["x"] == 1



from typing import Dict

from vyper.semantics.analysis.base import Modifiability, VarInfo
from vyper.semantics.types import AddressT, BytesT, SelfT, VyperType
from vyper.semantics.types.shortcuts import BYTES32_T, UINT256_T


# common properties for environment variables
class _EnvType(VyperType):
    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return hash(id(self))


class _Block(_EnvType):
    _id = "block"
    _type_members = {
        "coinbase": AddressT(),
        "difficulty": UINT256_T,
        "prevrandao": BYTES32_T,
        "number": UINT256_T,
        "gaslimit": UINT256_T,
        "basefee": UINT256_T,
        "blobbasefee": UINT256_T,
        "prevhash": BYTES32_T,
        "timestamp": UINT256_T,
    }


class _Chain(_EnvType):
    _id = "chain"
    _type_members = {"id": UINT256_T}


class _Msg(_EnvType):
    _id = "msg"
    _type_members = {"data": BytesT(), "gas": UINT256_T, "sender": AddressT(), "value": UINT256_T}


class _Tx(_EnvType):
    _id = "tx"
    _type_members = {"origin": AddressT(), "gasprice": UINT256_T}


CONSTANT_ENVIRONMENT_VARS = {t._id: t for t in (_Block(), _Chain(), _Tx(), _Msg())}


def get_constant_vars() -> Dict:
    """
    Get a dictionary of constant environment variables.
    """
    result = {}
    for k, v in CONSTANT_ENVIRONMENT_VARS.items():
        result[k] = VarInfo(v, modifiability=Modifiability.RUNTIME_CONSTANT)

    return result


MUTABLE_ENVIRONMENT_VARS: Dict[str, type] = {"self": SelfT}


def get_mutable_vars() -> Dict:
    """
    Get a dictionary of mutable environment variables (those that are
    modified during the course of contract execution, such as `self`).
    """
    return {name: VarInfo(type_()) for name, type_ in MUTABLE_ENVIRONMENT_VARS.items()}



import pytest

from vyper import compiler
from vyper.exceptions import StructureException, UndeclaredDefinition, UnknownAttribute

fail_list = [
    (
        """
@external
def foo() -> uint256:
    doesnotexist(2, uint256)
    return convert(2, uint256)
    """,
        UndeclaredDefinition,
    ),
    (
        """
@external
def foo(x: int256) -> uint256:
    convert(x, uint256)
    return convert(x, uint256)

    """,
        StructureException,
    ),
    (
        """
@internal
def test(a : uint256):
    pass


@external
def burn(_value: uint256):
    self.test(msg.sender._value)
    """,
        UnknownAttribute,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_functions_call_fail(bad_code, exc):
    with pytest.raises(exc):
        compiler.compile_code(bad_code)


valid_list = [
    """
@external
def foo(x: int128) -> uint256:
    return convert(x, uint256)
    """,
    """
from ethereum.ercs import IERC20

interface Factory:
    def getExchange(token_addr: address) -> address: view

token: IERC20
factory: Factory

@external
def setup(token_addr: address):
    self.token = IERC20(token_addr)
    assert staticcall self.factory.getExchange(self.token.address) == self
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_functions_call_success(good_code):
    assert compiler.compile_code(good_code) is not None



from vyper import ast as vy_ast
from vyper.abi_types import ABI_Bytes, ABI_String, ABIType
from vyper.exceptions import CompilerPanic, StructureException, UnexpectedNodeType, UnexpectedValue
from vyper.semantics.types.base import VyperType
from vyper.semantics.types.utils import get_index_value
from vyper.utils import ceil32


class _BytestringT(VyperType):
    """
    Private base class for single-value types which occupy multiple memory slots
    and where a maximum length must be given via a subscript (string, bytes).

    Types for literals have an inferred minimum length. For example, `b"hello"`
    has a length of 5 of more and so can be used in an operation with `bytes[5]`
    or `bytes[10]`, but not `bytes[4]`. Upon comparison to a fixed length type,
    the minimum length is discarded and the type assumes the fixed length it was
    compared against.

    Attributes
    ----------
    _length : int
        The maximum allowable length of the data within the type.
    _min_length: int
        The minimum length of the data within the type. Used when the type
        is applied to a literal definition.
    """

    # this is a carveout because currently we allow dynamic arrays of
    # bytestrings, but not static arrays of bytestrings
    _as_darray = True
    _as_hashmap_key = True
    _equality_attrs = ("_length", "_min_length")
    _is_bytestring: bool = True

    def __init__(self, length: int = 0) -> None:
        super().__init__()

        self._length = length
        self._min_length = length

    def __repr__(self):
        return f"{self._id}[{self.length}]"

    def _addl_dict_fields(self):
        return {"length": self.length}

    @property
    def length(self):
        """
        Property method used to check the length of a type.
        """
        if self._length:
            return self._length
        return self._min_length

    @property
    def maxlen(self):
        """
        Alias for backwards compatibility.
        """
        return self.length

    def validate_literal(self, node: vy_ast.Constant) -> None:
        super().validate_literal(node)

        if len(node.value) != self.length:
            # should always be constructed with correct length
            # at the point that validate_literal is called
            raise CompilerPanic("unreachable")

    @property
    def size_in_bytes(self):
        # the first slot (32 bytes) stores the actual length, and then we reserve
        # enough additional slots to store the data if it uses the max available length
        # because this data type is single-bytes, we make it so it takes the max 32 byte
        # boundary as it's size, instead of giving it a size that is not cleanly divisible by 32

        return 32 + ceil32(self.length)

    def set_length(self, length):
        """
        Sets the exact length of the type.

        May only be called once, and only on a type that does not yet have
        a fixed length.
        """
        if self._length:
            raise CompilerPanic("Type already has a fixed length")
        self._length = length
        self._min_length = length

    def set_min_length(self, min_length):
        """
        Sets the minimum length of the type.

        May only be used to increase the minimum length. May not be called if
        an exact length has been set.
        """
        if self._length:
            raise CompilerPanic("Type already has a fixed length")
        if self._min_length > min_length:
            raise CompilerPanic("Cannot reduce the min_length of ArrayValueType")
        self._min_length = min_length

    def compare_type(self, other):
        if not super().compare_type(other):
            return False

        # CMC 2022-03-18 TODO this method should be refactored so it does not have side effects

        # when comparing two literals, both now have an equal min-length
        if not self._length and not other._length:
            min_length = max(self._min_length, other._min_length)
            self.set_min_length(min_length)
            other.set_min_length(min_length)
            return True

        # comparing a defined length to a literal causes the literal to have a fixed length
        if self._length:
            if not other._length:
                other.set_length(max(self._length, other._min_length))
            return self._length >= other._length

        return other.compare_type(self)

    @classmethod
    def from_annotation(cls, node: vy_ast.VyperNode) -> "_BytestringT":
        if not isinstance(node, vy_ast.Subscript):
            raise StructureException(
                f"Cannot declare {cls._id} type without a maximum length, e.g. {cls._id}[5]", node
            )

        if node.get("value.id") != cls._id:
            raise UnexpectedValue("Node id does not match type name")

        length = get_index_value(node.slice)  # type: ignore

        if length is None:
            raise StructureException(
                f"Cannot declare {cls._id} type without a maximum length, e.g. {cls._id}[5]", node
            )

        # TODO: pass None to constructor after we redo length inference on bytestrings
        length = length or 0

        return cls(length)

    @classmethod
    def from_literal(cls, node: vy_ast.Constant) -> "_BytestringT":
        if not isinstance(node, cls._valid_literal):
            raise UnexpectedNodeType(f"Not a {cls._id}: {node}")
        t = cls()
        t.set_min_length(len(node.value))
        return t


class BytesT(_BytestringT):
    typeclass = "bytes"

    _id = "Bytes"
    _valid_literal = (vy_ast.Bytes,)

    @property
    def abi_type(self) -> ABIType:
        return ABI_Bytes(self.length)


class StringT(_BytestringT):
    typeclass = "string"

    _id = "String"
    _valid_literal = (vy_ast.Str,)

    @property
    def abi_type(self) -> ABIType:
        return ABI_String(self.length)



import operator
from typing import Callable

from vyper.utils import (
    SizeLimits,
    evm_div,
    evm_mod,
    evm_pow,
    signed_to_unsigned,
    unsigned_to_signed,
)
from vyper.venom.basicblock import IROperand


def _unsigned_to_signed(value: int) -> int:
    if value <= SizeLimits.MAX_INT256:
        return value  # fast exit
    else:
        return unsigned_to_signed(value, 256)


def _signed_to_unsigned(value: int) -> int:
    if value >= 0:
        return value  # fast exit
    else:
        return signed_to_unsigned(value, 256)


def _wrap_signed_binop(operation):
    def wrapper(ops: list[IROperand]) -> int:
        first = _unsigned_to_signed(ops[1].value)
        second = _unsigned_to_signed(ops[0].value)
        return _signed_to_unsigned(int(operation(first, second)))

    return wrapper


def _wrap_binop(operation):
    def wrapper(ops: list[IROperand]) -> int:
        first = _signed_to_unsigned(ops[1].value)
        second = _signed_to_unsigned(ops[0].value)
        ret = operation(first, second)
        assert isinstance(ret, int)
        return ret & SizeLimits.MAX_UINT256

    return wrapper


def _evm_signextend(ops: list[IROperand]) -> int:
    value = ops[0].value
    nbytes = ops[1].value

    assert 0 <= value <= SizeLimits.MAX_UINT256, "Value out of bounds"

    if nbytes > 31:
        return value

    sign_bit = 1 << (nbytes * 8 + 7)
    if value & sign_bit:
        value |= SizeLimits.CEILING_UINT256 - sign_bit
    else:
        value &= sign_bit - 1

    return value


def _evm_iszero(ops: list[IROperand]) -> int:
    value = ops[0].value
    assert SizeLimits.MIN_INT256 <= value <= SizeLimits.MAX_UINT256, "Value out of bounds"
    return int(value == 0)  # 1 if True else 0


def _evm_shr(ops: list[IROperand]) -> int:
    value = ops[0].value
    shift_len = ops[1].value
    assert 0 <= value <= SizeLimits.MAX_UINT256, "Value out of bounds"
    return value >> shift_len


def _evm_shl(ops: list[IROperand]) -> int:
    value = ops[0].value
    shift_len = ops[1].value
    assert 0 <= value <= SizeLimits.MAX_UINT256, "Value out of bounds"
    if shift_len >= 256:
        return 0
    return (value << shift_len) & SizeLimits.MAX_UINT256


def _evm_sar(ops: list[IROperand]) -> int:
    value = _unsigned_to_signed(ops[0].value)
    assert SizeLimits.MIN_INT256 <= value <= SizeLimits.MAX_INT256, "Value out of bounds"
    shift_len = ops[1].value
    return value >> shift_len


def _evm_not(ops: list[IROperand]) -> int:
    value = ops[0].value
    assert 0 <= value <= SizeLimits.MAX_UINT256, "Value out of bounds"
    return SizeLimits.MAX_UINT256 ^ value


ARITHMETIC_OPS: dict[str, Callable[[list[IROperand]], int]] = {
    "add": _wrap_binop(operator.add),
    "sub": _wrap_binop(operator.sub),
    "mul": _wrap_binop(operator.mul),
    "div": _wrap_binop(evm_div),
    "sdiv": _wrap_signed_binop(evm_div),
    "mod": _wrap_binop(evm_mod),
    "smod": _wrap_signed_binop(evm_mod),
    "exp": _wrap_binop(evm_pow),
    "eq": _wrap_binop(operator.eq),
    "ne": _wrap_binop(operator.ne),
    "lt": _wrap_binop(operator.lt),
    "le": _wrap_binop(operator.le),
    "gt": _wrap_binop(operator.gt),
    "ge": _wrap_binop(operator.ge),
    "slt": _wrap_signed_binop(operator.lt),
    "sle": _wrap_signed_binop(operator.le),
    "sgt": _wrap_signed_binop(operator.gt),
    "sge": _wrap_signed_binop(operator.ge),
    "or": _wrap_binop(operator.or_),
    "and": _wrap_binop(operator.and_),
    "xor": _wrap_binop(operator.xor),
    "not": _evm_not,
    "signextend": _evm_signextend,
    "iszero": _evm_iszero,
    "shr": _evm_shr,
    "shl": _evm_shl,
    "sar": _evm_sar,
    "store": lambda ops: ops[0].value,
}



import pytest

from vyper import compiler
from vyper.exceptions import TypeMismatch

fail_list = [
    (
        """
@external
def foo():
    send(1, 2)
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo():
    send(0x1234567890123456789012345678901234567890, 2.5)
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo():
    send(0x1234567890123456789012345678901234567890, 0x1234567890123456789012345678901234567890)
    """,
        TypeMismatch,
    ),
    (
        """
x: int128

@external
def foo():
    send(0x1234567890123456789012345678901234567890, self.x)
    """,
        TypeMismatch,
    ),
    (
        """
x: uint256

@external
def foo():
    send(0x1234567890123456789012345678901234567890, self.x + 1.5)
    """,
        TypeMismatch,
    ),
    (
        """
x: decimal

@external
def foo():
    send(0x1234567890123456789012345678901234567890, self.x)
    """,
        TypeMismatch,
    ),
    # Tests for sending gas stipend
    (
        """
@external
def foo():
    send(0x1234567890123456789012345678901234567890, 5, gas=1.5)
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo():
    send(0x1234567890123456789012345678901234567890, 5, gas=-2)
    """,
        TypeMismatch,
    ),
    (
        """
x: int128

@external
def foo():
    send(0x1234567890123456789012345678901234567890, 5, gas=self.x)
    """,
        TypeMismatch,
    ),
    (
        """
x: decimal

@external
def foo():
    send(0x1234567890123456789012345678901234567890, 5, gas=self.x)
    """,
        TypeMismatch,
    ),
    # End tests for sending gas stipend
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_send_fail(bad_code, exc):
    with pytest.raises(exc):
        compiler.compile_code(bad_code)


valid_list = [
    """
x: uint256

@external
def foo():
    send(0x1234567890123456789012345678901234567890, self.x + 1)
    """,
    """
x: decimal

@external
def foo():
    send(
        0x1234567890123456789012345678901234567890,
        as_wei_value(convert(floor(self.x), int128),
        "wei")
    )
    """,
    """
x: uint256

@external
def foo():
    send(0x1234567890123456789012345678901234567890, self.x)
    """,
    """
@external
def foo():
    send(0xde0B295669a9FD93d5F28D9Ec85E40f4cb697BAe, 5)
    """,
    """
# Test custom send method
@internal
def send(a: address, w: uint256):
    send(0xde0B295669a9FD93d5F28D9Ec85E40f4cb697BAe, 1)

@external
@payable
def foo():
    self.send(msg.sender, msg.value)
    """,
    """
#Test send gas stipend
@external
def foo():
    send(0xde0B295669a9FD93d5F28D9Ec85E40f4cb697BAe, 5, gas=5000)
    """,
    """
x: uint256

#Test send gas stipend
@external
def foo():
    send(0xde0B295669a9FD93d5F28D9Ec85E40f4cb697BAe, 5, gas=self.x)
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_block_success(good_code):
    assert compiler.compile_code(good_code) is not None



import pytest

from tests.utils import parse_and_fold


@pytest.mark.parametrize("length", [0, 1, 32, 33, 64, 65, 1024])
def test_len_string(get_contract, length):
    source = """
@external
def foo(a: String[1024]) -> uint256:
    return len(a)
    """
    contract = get_contract(source)

    value = "a" * length

    vyper_ast = parse_and_fold(f"len('{value}')")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(value) == new_node.value


@pytest.mark.parametrize("length", [0, 1, 32, 33, 64, 65, 1024])
def test_len_bytes(get_contract, length):
    source = """
@external
def foo(a: Bytes[1024]) -> uint256:
    return len(a)
    """
    contract = get_contract(source)

    value = "a" * length

    vyper_ast = parse_and_fold(f"len(b'{value}')")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(value.encode()) == new_node.value


@pytest.mark.parametrize("length", [1, 32, 33, 64, 65, 1024])
def test_len_hex(get_contract, length):
    source = """
@external
def foo(a: Bytes[1024]) -> uint256:
    return len(a)
    """
    contract = get_contract(source)

    value = f"0x{'00' * length}"

    vyper_ast = parse_and_fold(f"len({value})")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(value) == new_node.value



import re

import pytest

from tests.utils import decimal_to_int
from vyper.compiler import compile_code
from vyper.exceptions import (
    ArgumentException,
    ImmutableViolation,
    InvalidType,
    IteratorException,
    NamespaceCollision,
    StructureException,
    SyntaxException,
    TypeMismatch,
    UnknownType,
)

BASIC_FOR_LOOP_CODE = [
    # basic for-in-list memory
    (
        """
@external
def data() -> int128:
    s: int128[5] = [1, 2, 3, 4, 5]
    for i: int128 in s:
        if i >= 3:
            return i
    return -1""",
        3,
    ),
    # basic for-in-dynamic array
    (
        """
@external
def data() -> int128:
    s: DynArray[int128, 10] = [1, 2, 3, 4, 5]
    for i: int128 in s:
        if i >= 3:
            return i
    return -1""",
        3,
    ),
    # test more complicated type
    (
        """
struct S:
    x: int128
    y: int128

@external
def data() -> int128:
    sss: DynArray[DynArray[S, 10], 10] = [
        [S(x=1, y=2)],
        [S(x=3, y=4), S(x=5, y=6), S(x=7, y=8), S(x=9, y=10)]
        ]
    ret: int128 = 0
    for ss: DynArray[S, 10] in sss:
        for s: S in ss:
            ret += s.x + s.y
    return ret""",
        sum(range(1, 11)),
    ),
    # basic for-in-list literal
    (
        """
@external
def data() -> int128:
    for i: int128 in [3, 5, 7, 9]:
        if i > 5:
            return i
    return -1""",
        7,
    ),
    (
        # test variable string dynarray
        """
@external
def data() -> String[33]:
    xs: DynArray[String[33], 3] = ["hello", ",", "world"]
    for x: String[33] in xs:
        if x == ",":
            return x
    return ""
    """,
        ",",
    ),
    (
        # test literal string dynarray
        """
@external
def data() -> String[33]:
    for x: String[33] in ["hello", ",", "world"]:
        if x == ",":
            return x
    return ""
    """,
        ",",
    ),
    (
        # test nested string dynarray
        """
@external
def data() -> DynArray[String[33], 2]:
    for x: DynArray[String[33], 2] in [["hello", "world"], ["goodbye", "world!"]]:
        if x[1] == "world":
            return x
    return []
    """,
        ["hello", "world"],
    ),
    # test nested array
    (
        """
@external
def data() -> int128:
    ret: int128 = 0
    xss: int128[3][3] = [[1,2,3],[4,5,6],[7,8,9]]
    for xs: int128[3] in xss:
        for x: int128 in xs:
            ret += x
    return ret""",
        sum(range(1, 10)),
    ),
    # test more complicated literal
    (
        """
struct S:
    x: int128
    y: int128

@external
def data() -> int128:
    ret: int128 = 0
    for ss: S[1] in [[S(x=1, y=2)]]:
        for s: S in ss:
            ret += s.x + s.y
    return ret""",
        1 + 2,
    ),
    # basic for-in-list addresses
    (
        """
@external
def data() -> address:
    addresses: address[3] = [
        0x7d577a597B2742b498Cb5Cf0C26cDCD726d39E6e,
        0x82A978B3f5962A5b0957d9ee9eEf472EE55B42F1,
        0xDCEceAF3fc5C0a63d195d69b1A90011B7B19650D
    ]
    count: int128 = 0
    for i: address in addresses:
        count += 1
        if count == 2:
            return i
    return 0x0000000000000000000000000000000000000000
    """,
        "0x82A978B3f5962A5b0957d9ee9eEf472EE55B42F1",
    ),
]


@pytest.mark.parametrize("code, data", BASIC_FOR_LOOP_CODE)
def test_basic_for_in_lists(code, data, get_contract):
    c = get_contract(code)
    assert c.data() == data


def test_basic_for_list_storage(get_contract):
    code = """
x: int128[4]

@external
def set():
    self.x = [3, 5, 7, 9]

@external
def data() -> int128:
    for i: int128 in self.x:
        if i > 5:
            return i
    return -1
    """

    c = get_contract(code)

    assert c.data() == -1
    c.set()
    assert c.data() == 7


def test_basic_for_dyn_array_storage(get_contract):
    code = """
x: DynArray[int128, 4]

@external
def set(xs: DynArray[int128, 4]):
    self.x = xs

@external
def data() -> int128:
    t: int128 = 0
    for i: int128 in self.x:
        t += i
    return t
    """

    c = get_contract(code)

    assert c.data() == 0
    # test all sorts of lists
    for xs in [[3, 5, 7, 9], [4, 6, 8], [1, 2], [5], []]:
        c.set(xs)
        assert c.data() == sum(xs)


def test_basic_for_list_storage_address(get_contract):
    code = """
addresses: address[3]

@external
def set(i: int128, val: address):
    self.addresses[i] = val

@external
def ret(i: int128) -> address:
    return self.addresses[i]

@external
def iterate_return_second() -> address:
    count: int128 = 0
    for i: address in self.addresses:
        count += 1
        if count == 2:
            return i
    return empty(address)
    """

    c = get_contract(code)

    c.set(0, "0x82A978B3f5962A5b0957d9ee9eEf472EE55B42F1")
    c.set(1, "0x7d577a597B2742b498Cb5Cf0C26cDCD726d39E6e")
    c.set(2, "0xDCEceAF3fc5C0a63d195d69b1A90011B7B19650D")

    assert c.ret(1) == c.iterate_return_second() == "0x7d577a597B2742b498Cb5Cf0C26cDCD726d39E6e"


def test_basic_for_list_storage_decimal(get_contract):
    code = """
readings: decimal[3]

@external
def set(i: int128, val: decimal):
    self.readings[i] = val

@external
def ret(i: int128) -> decimal:
    return self.readings[i]

@external
def i_return(break_count: int128) -> decimal:
    count: int128 = 0
    for i: decimal in self.readings:
        if count == break_count:
            return i
        count += 1
    return -1.111
    """

    c = get_contract(code)

    c.set(0, decimal_to_int("0.0001"))
    c.set(1, decimal_to_int("1.1"))
    c.set(2, decimal_to_int("2.2"))

    assert c.ret(2) == c.i_return(2) == decimal_to_int("2.2")
    assert c.ret(1) == c.i_return(1) == decimal_to_int("1.1")
    assert c.ret(0) == c.i_return(0) == decimal_to_int("0.0001")


def test_for_in_list_iter_type(get_contract):
    code = """
@external
@view
def func(amounts: uint256[3]) -> uint256:
    total: uint256 = as_wei_value(0, "wei")

    # calculate total
    for amount: uint256 in amounts:
        total += amount

    return total
    """

    c = get_contract(code)

    assert c.func([100, 200, 300]) == 600


def test_for_in_dyn_array(get_contract):
    code = """
@external
@view
def func(amounts: DynArray[uint256, 3]) -> uint256:
    total: uint256 = 0

    # calculate total
    for amount: uint256 in amounts:
        total += amount

    return total
    """

    c = get_contract(code)

    assert c.func([100, 200, 300]) == 600
    assert c.func([100, 200]) == 300


GOOD_CODE = [
    # multiple for loops
    """
@external
def foo(x: int128):
    p: int128 = 0
    for i: int128 in range(3):
        p += i
    for i: int128 in range(4):
        p += i
    """,
    """
@external
def foo(x: int128):
    p: int128 = 0
    for i: int128 in range(3):
        p += i
    for i: int128 in [1, 2, 3, 4]:
        p += i
    """,
    """
@external
def foo(x: int128):
    p: int128 = 0
    for i: int128 in [1, 2, 3, 4]:
        p += i
    for i: int128 in [1, 2, 3, 4]:
        p += i
    """,
    """
@external
def foo():
    for i: uint256 in range(10):
        pass
    for i: uint256 in range(20):
        pass
    """,
    # using index variable after loop
    """
@external
def foo():
    for i: uint256 in range(10):
        pass
    i: int128 = 100  # create new variable i
    i = 200  # look up the variable i and check whether it is in forvars
    """,
]


@pytest.mark.parametrize("code", GOOD_CODE)
def test_good_code(code, get_contract):
    get_contract(code)


RANGE_CONSTANT_CODE = [
    (
        """
TREE_FIDDY: constant(uint256)  = 350


@external
def a() -> uint256:
    x: uint256 = 0
    for i: uint256 in range(TREE_FIDDY):
        x += 1
    return x""",
        350,
    ),
    (
        """
ONE_HUNDRED: constant(uint256)  = 100

@external
def a() -> uint256:
    x: uint256 = 0
    for i: uint256 in range(1, 1 + ONE_HUNDRED):
        x += 1
    return x""",
        100,
    ),
    (
        """
START: constant(int128)  = 100
END: constant(int128)  = 199

@external
def a() -> int128:
    x: int128 = 0
    for i: int128 in range(START, END):
        x += 1
    return x""",
        99,
    ),
    (
        """
@external
def a() -> int128:
    x: int128 = 0
    for i: int128 in range(-5, -1):
        x += i
    return x""",
        -14,
    ),
    (
        """
@external
def a() -> uint256:
    a: DynArray[DynArray[uint256, 2], 3] = [[0, 1], [2, 3], [4, 5]]
    x: uint256 = 0
    for i: uint256 in a[2]:
        x += i
    return x
    """,
        9,
    ),
]


@pytest.mark.parametrize("code, result", RANGE_CONSTANT_CODE)
def test_range_constant(get_contract, code, result):
    c = get_contract(code)

    assert c.a() == result


BAD_CODE = [
    # altering list within loop
    (
        """
@external
def data() -> int128:
    s: int128[6] = [1, 2, 3, 4, 5, 6]
    count: int128 = 0
    for i: int128 in s:
        s[count] = 1  # this should not be allowed.
        if i >= 3:
            return i
        count += 1
    return -1
    """,
        ImmutableViolation,
    ),
    (
        """
@external
def foo():
    s: int128[6] = [1, 2, 3, 4, 5, 6]
    count: int128 = 0
    for i: int128 in s:
        s[count] += 1
    """,
        ImmutableViolation,
    ),
    # alter storage list within for loop
    (
        """
s: int128[6]

@external
def set():
    self.s = [1, 2, 3, 4, 5, 6]

@external
def data() -> int128:
    count: int128 = 0
    for i: int128 in self.s:
        self.s[count] = 1  # this should not be allowed.
        if i >= 3:
            return i
        count += 1
    return -1
    """,
        ImmutableViolation,
    ),
    # alter nested storage list in internal function call within for loop
    (
        """
struct Foo:
    foo: uint256[4]

my_array2: Foo

@internal
def doStuff(i: uint256) -> uint256:
    self.my_array2.foo[i] = i
    return i

@internal
def _helper():
    i: uint256 = 0
    for item: uint256 in self.my_array2.foo:
        self.doStuff(i)
        i += 1
    """,
        ImmutableViolation,
    ),
    # alter doubly nested storage list in internal function call within for loop
    (
        """
struct Foo:
    foo: uint256[4]

struct Bar:
    bar: Foo
    baz: uint256

my_array2: Bar

@internal
def doStuff(i: uint256) -> uint256:
    self.my_array2.bar.foo[i] = i
    return i

@internal
def _helper():
    i: uint256 = 0
    for item: uint256 in self.my_array2.bar.foo:
        self.doStuff(i)
        i += 1
    """,
        ImmutableViolation,
    ),
    # alter entire struct with nested storage list in internal function call within for loop
    (
        """
struct Foo:
    foo: uint256[4]

my_array2: Foo

@internal
def doStuff():
    self.my_array2.foo = [
        block.timestamp + 1,
        block.timestamp + 2,
        block.timestamp + 3,
        block.timestamp + 4
    ]

@internal
def _helper():
    i: uint256 = 0
    for item: uint256 in self.my_array2.foo:
        self.doStuff()
        i += 1
    """,
        ImmutableViolation,
    ),
    # invalid nested loop
    (
        """
@external
def foo(x: int128):
    for i: int128 in range(4):
        for i: int128 in range(5):
            pass
    """,
        NamespaceCollision,
    ),
    (
        """
@external
def foo(x: int128):
    for i: int128 in [1,2]:
        for i: int128 in [1,2]:
            pass
     """,
        NamespaceCollision,
    ),
    # invalid iterator assignment
    (
        """
@external
def foo(x: int128):
    for i: int128 in [1,2]:
        i = 2
    """,
        ImmutableViolation,
    ),
    # invalid modification of dynarray
    (
        """
@external
def foo():
    xs: DynArray[uint256, 5] = [1,2,3]
    for x: uint256 in xs:
        xs.pop()
    """,
        ImmutableViolation,
    ),
    # invalid modification of dynarray
    (
        """
@external
def foo():
    xs: DynArray[uint256, 5] = [1,2,3]
    for x: uint256 in xs:
        xs.append(x)
    """,
        ImmutableViolation,
    ),
    # invalid modification of dynarray
    (
        """
@external
def foo():
    xs: DynArray[DynArray[uint256, 5], 5] = [[1,2,3]]
    for x: DynArray[uint256, 5] in xs:
        x.pop()
    """,
        ImmutableViolation,
    ),
    # invalid modification of dynarray
    (
        """
array: DynArray[uint256, 5]
@internal
def a():
    self.b()

@internal
def b():
    self.array.pop()

@external
def foo():
    for x: uint256 in self.array:
        self.a()
    """,
        ImmutableViolation,
    ),
    (
        """
@external
def foo(x: int128):
    for i: int128 in [1,2]:
        i += 2
    """,
        ImmutableViolation,
    ),
    # range of < 1
    (
        """
@external
def foo():
    for i: int128 in range(-3):
        pass
    """,
        StructureException,
    ),
    """
@external
def foo():
    for i: uint256 in range(0):
        pass
    """,
    """
@external
def foo():
    for i: uint256 in []:
        pass
    """,
    """
FOO: constant(DynArray[uint256, 3]) = []

@external
def foo():
    for i: uint256 in FOO:
        pass
    """,
    (
        """
@external
def foo():
    for i: uint256 in range(5,3):
        pass
    """,
        StructureException,
    ),
    (
        """
@external
def foo():
    for i: int128 in range(5,3,-1):
        pass
    """,
        ArgumentException,
    ),
    (
        """
@external
def foo():
    a: uint256 = 2
    for i: uint256 in range(a):
        pass
    """,
        StructureException,
    ),
    (
        """
@external
def foo():
    a: int128 = 6
    for i: int128 in range(a,a-3):
        pass
    """,
        StructureException,
    ),
    # invalid argument length
    (
        """
@external
def foo():
    for i: uint256 in range():
        pass
    """,
        ArgumentException,
    ),
    (
        """
@external
def foo():
    for i: uint256 in range(0,1,2):
        pass
    """,
        ArgumentException,
    ),
    # non-iterables
    (
        """
@external
def foo():
    for i: Bytes[1] in b"asdf":
        pass
    """,
        InvalidType,
    ),
    (
        """
@external
def foo():
    for i: uint256 in 31337:
        pass
    """,
        InvalidType,
    ),
    (
        """
@external
def foo():
    for i: uint256 in bar():
        pass
    """,
        IteratorException,
    ),
    (
        """
@external
def foo():
    for i: uint256 in self.bar():
        pass
    """,
        IteratorException,
    ),
    (
        """
@external
def test_for() -> int128:
    a: int128 = 0
    for i: int128 in range(max_value(int128), max_value(int128)+2):
        a = i
    return a
    """,
        TypeMismatch,
    ),
    (
        """
@external
def test_for() -> int128:
    a: int128 = 0
    b: uint256 = 0
    for i: int128 in range(5):
        a = i
        b = i
    return a
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo():
    for i in [1, 2, 3]:
        pass
    """,
        SyntaxException,
    ),
    (
        """
@external
def foo():
    for i: $$$ in [1, 2, 3]:
        pass
    """,
        SyntaxException,
    ),
    (
        """
@external
def foo():
    for i: uint9 in [1, 2, 3]:
        pass
    """,
        UnknownType,
    ),
]

BAD_CODE = [code if isinstance(code, tuple) else (code, StructureException) for code in BAD_CODE]
for_code_regex = re.compile(r"for .+ in (.*):")
bad_code_names = [
    f"{i} {for_code_regex.search(code).group(1)}" for i, (code, _) in enumerate(BAD_CODE)
]


# TODO: move these to tests/functional/syntax
@pytest.mark.parametrize("code,err", BAD_CODE, ids=bad_code_names)
def test_bad_code(assert_compile_failed, get_contract, code, err):
    with pytest.raises(err):
        compile_code(code)


def test_iterator_modification_module_attribute(make_input_bundle):
    # test modifying iterator via attribute
    lib1 = """
queue: DynArray[uint256, 5]
    """
    main = """
import lib1

initializes: lib1

@external
def foo():
    for i: uint256 in lib1.queue:
        lib1.queue.pop()
    """

    input_bundle = make_input_bundle({"lib1.vy": lib1})

    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot modify loop variable `queue`"


def test_iterator_modification_module_function_call(make_input_bundle):
    lib1 = """
queue: DynArray[uint256, 5]

@internal
def popqueue():
    self.queue.pop()
    """
    main = """
import lib1

initializes: lib1

@external
def foo():
    for i: uint256 in lib1.queue:
        lib1.popqueue()
    """

    input_bundle = make_input_bundle({"lib1.vy": lib1})

    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot modify loop variable `queue`"



import vyper
from vyper.compiler.settings import Settings

source = """
@external
def foo(a: uint256):
    _numBids: uint256 = 20
    b: uint256 = 10

    for i: uint256 in range(128):
        b = 1 + _numBids
"""


def test_liveness_simple_loop():
    vyper.compile_code(
        source, output_formats=["opcodes"], settings=Settings(experimental_codegen=True)
    )



import pytest

from vyper import compiler
from vyper.exceptions import InvalidReference

fail_list = [
    """
x: uint256

@external
def foo():
    send(0x1234567890123456789012345678901234567890, x)
    """,
    """
@external
def bar(x: int128) -> int128:
    return 3 * x

@external
def foo() -> int128:
    return bar(20)
    """,
    """
b: int128
@external
def foo():
    b = 7
    """,
    """
x: int128
@external
def foo():
    x = 5
    """,
    """
@external
def foo():
    int128 = 5
    """,
    """
a: public(constant(uint256)) = 1

@external
def foo():
    b: uint256 = self.a
    """,
    """
a: public(immutable(uint256))

@deploy
def __init__():
    a = 123

@external
def foo():
    b: uint256 = self.a
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_invalid_reference_exception(bad_code):
    with pytest.raises(InvalidReference):
        compiler.compile_code(bad_code)



from decimal import Decimal

import pytest
from hypothesis import example, given, settings
from hypothesis import strategies as st

from tests.utils import decimal_to_int, parse_and_fold
from vyper.exceptions import OverflowException, TypeMismatch, ZeroDivisionException
from vyper.semantics.analysis.local import ExprVisitor
from vyper.semantics.types import DecimalT

DECIMAL_T = DecimalT()

st_decimals = st.decimals(
    min_value=DECIMAL_T.decimal_bounds[0],
    max_value=DECIMAL_T.decimal_bounds[1],
    allow_nan=False,
    allow_infinity=False,
    places=DECIMAL_T._decimal_places,
)


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(left=st_decimals, right=st_decimals)
@example(left=Decimal("0.9999999999"), right=Decimal("0.0000000001"))
@example(left=Decimal("0.0000000001"), right=Decimal("0.9999999999"))
@example(left=Decimal("0.9999999999"), right=Decimal("0.9999999999"))
@example(left=Decimal("0.0000000001"), right=Decimal("0.0000000001"))
@pytest.mark.parametrize("op", "+-*/%")
def test_binop_decimal(get_contract, tx_failed, op, left, right):
    source = f"""
@external
def foo(a: decimal, b: decimal) -> decimal:
    return a {op} b
    """
    contract = get_contract(source)

    try:
        vyper_ast = parse_and_fold(f"{left} {op} {right}")
        expr = vyper_ast.body[0].value

        # check invalid values
        ExprVisitor().visit(expr, DecimalT())

        new_node = expr.get_folded_value()
        is_valid = True
    except (OverflowException, ZeroDivisionException):
        is_valid = False

    left = decimal_to_int(left)
    right = decimal_to_int(right)
    if is_valid:
        assert contract.foo(left, right) == decimal_to_int(new_node.value)
    else:
        with tx_failed():
            contract.foo(left, right)


def test_binop_pow():
    # raises because Vyper does not support decimal exponentiation
    with pytest.raises(TypeMismatch):
        _ = parse_and_fold("3.1337 ** 4.2")


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(
    values=st.lists(st_decimals, min_size=2, max_size=10),
    ops=st.lists(st.sampled_from("+-*/%"), min_size=11, max_size=11),
)
def test_nested(get_contract, tx_failed, values, ops):
    variables = "abcdefghij"
    input_value = ",".join(f"{i}: decimal" for i in variables[: len(values)])
    return_value = " ".join(f"{a} {b}" for a, b in zip(variables[: len(values)], ops))
    return_value = return_value.rsplit(maxsplit=1)[0]
    source = f"""
@external
def foo({input_value}) -> decimal:
    return {return_value}
    """
    contract = get_contract(source)

    literal_op = " ".join(f"{a} {b}" for a, b in zip(values, ops))
    literal_op = literal_op.rsplit(maxsplit=1)[0]
    try:
        vyper_ast = parse_and_fold(literal_op)
        expr = vyper_ast.body[0].value

        # check invalid intermediate values
        ExprVisitor().visit(expr, DecimalT())

        new_node = expr.get_folded_value()
        expected = new_node.value
        is_valid = True
    except (OverflowException, ZeroDivisionException):
        # for overflow or division/modulus by 0, expect the contract call to revert
        is_valid = False

    values = [decimal_to_int(v) for v in values]
    if is_valid:
        assert contract.foo(*values) == decimal_to_int(expected)
    else:
        with tx_failed():
            contract.foo(*values)



import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from tests.utils import decimal_to_int, parse_and_fold
from vyper.builtins import functions as vy_fn
from vyper.utils import SizeLimits

denoms = [x for k in vy_fn.AsWeiValue.wei_denoms.keys() for x in k]


st_decimals = st.decimals(
    min_value=0,
    max_value=SizeLimits.MAX_AST_DECIMAL,
    allow_nan=False,
    allow_infinity=False,
    places=10,
)


@pytest.mark.fuzzing
@settings(max_examples=10)
@given(value=st_decimals)
@pytest.mark.parametrize("denom", denoms)
def test_decimal(get_contract, value, denom):
    source = f"""
@external
def foo(a: decimal) -> uint256:
    return as_wei_value(a, '{denom}')
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"as_wei_value({value:.10f}, '{denom}')")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert isinstance(new_node.value, int)
    assert contract.foo(decimal_to_int(value)) == new_node.value


@pytest.mark.fuzzing
@settings(max_examples=10)
@given(value=st.integers(min_value=0, max_value=2**128))
@pytest.mark.parametrize("denom", denoms)
def test_integer(get_contract, value, denom):
    source = f"""
@external
def foo(a: uint256) -> uint256:
    return as_wei_value(a, '{denom}')
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"as_wei_value({value}, '{denom}')")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(value) == new_node.value



import pytest


@pytest.fixture(scope="module")
def c(env, get_contract):
    with open("examples/crowdfund.vy") as f:
        contract_code = f.read()
    return get_contract(contract_code, *[env.accounts[1], 50, 60])


def test_crowdfund_example(c, env):
    a0, a1, a2, a3, a4, a5, a6 = env.accounts[:7]
    env.set_balance(a0, 100)
    c.participate(value=5)
    env.timestamp += 1  # make sure auction has started

    assert c.timelimit() == 60
    assert c.deadline() - env.timestamp == 59
    assert env.timestamp < c.deadline()  # expired
    assert env.get_balance(c.address) < c.goal()  # not reached
    c.participate(value=49)
    # assert c.reached()
    pre_bal = env.get_balance(a1)
    env.timestamp += 100
    assert env.timestamp > c.deadline()  # expired
    c.finalize()
    post_bal = env.get_balance(a1)
    assert post_bal - pre_bal == 54


def test_crowdfund_example2(c, env, tx_failed):
    a0, a1, a2, a3, a4, a5, a6 = env.accounts[:7]
    for i, a in enumerate(env.accounts[3:7]):
        env.set_balance(a, i + 1)

    c.participate(value=1, sender=a3)
    c.participate(value=2, sender=a4)
    c.participate(value=3, sender=a5)
    c.participate(value=4, sender=a6)

    assert c.timelimit() == 60
    env.timestamp += 100
    # assert c.expired()
    # assert not c.reached()
    pre_bals = [env.get_balance(x) for x in [a3, a4, a5, a6]]
    with tx_failed():
        c.refund(sender=a0)
    c.refund(sender=a3)
    with tx_failed():
        c.refund(sender=a3)
    c.refund(sender=a4)
    c.refund(sender=a5)
    c.refund(sender=a6)
    post_bals = [env.get_balance(x) for x in [a3, a4, a5, a6]]
    assert [y - x for x, y in zip(pre_bals, post_bals)] == [1, 2, 3, 4]



import warnings
from decimal import getcontext

import pytest

import vyper.compiler.settings as compiler_settings
from tests.utils import decimal_to_int
from vyper import compile_code
from vyper.exceptions import (
    DecimalOverrideException,
    FeatureException,
    InvalidOperation,
    OverflowException,
    TypeMismatch,
)
from vyper.utils import DECIMAL_EPSILON, SizeLimits, quantize


def test_decimal_override():
    getcontext().prec = 78  # setting prec to 78 is ok

    # consumers of vyper, even as a library, are not allowed to reduce Decimal precision
    with pytest.raises(DecimalOverrideException):
        getcontext().prec = 77

    with warnings.catch_warnings(record=True) as w:
        getcontext().prec = 79
        # check warnings were issued
        assert len(w) == 1
        assert (
            str(w[-1].message) == "Changing decimals precision could have unintended side effects!"
        )


@pytest.mark.parametrize("op", ["//", "**", "&", "|", "^"])
def test_invalid_ops(op):
    code = f"""
@external
def foo(x: decimal, y: decimal) -> decimal:
    return x {op} y
    """
    with pytest.raises(InvalidOperation):
        compile_code(code)


@pytest.mark.parametrize("op", ["not"])
def test_invalid_unary_ops(op):
    code = f"""
@external
def foo(x: decimal) -> decimal:
    return {op} x
    """
    with pytest.raises(InvalidOperation):
        compile_code(code)


def test_decimal_test(get_contract):
    decimal_test = """
@external
def foo() -> int256:
    return(floor(999.0))

@external
def fop() -> int256:
    return(floor(333.0 + 666.0))

@external
def foq() -> int256:
    return(floor(1332.1 - 333.1))

@external
def bar() -> int256:
    return(floor(27.0 * 37.0))

@external
def baz() -> int256:
    x: decimal = 27.0
    return(floor(x * 37.0))

@external
def mok() -> int256:
    return(floor(999999.0 / 7.0 / 11.0 / 13.0))

@external
def mol() -> int256:
    return(floor(499.5 / 0.5))

@external
def mom() -> int256:
    return(floor(1498.5 / 1.5))

@external
def moo() -> int256:
    return(floor(2997.0 / 3.0))

@external
def foom() -> int256:
    return(floor(1999.0 % 1000.0))

@external
def foop() -> int256:
    return(floor(1999.0 % 1000.0))
    """

    c = get_contract(decimal_test)

    assert c.foo() == 999
    assert c.fop() == 999
    assert c.foq() == 999
    assert c.bar() == 999
    assert c.baz() == 999
    assert c.mok() == 999
    assert c.mol() == 999
    assert c.mom() == 999
    assert c.moo() == 999
    assert c.foom() == 999
    assert c.foop() == 999

    print("Passed basic addition, subtraction and multiplication tests")


def test_harder_decimal_test(get_contract):
    harder_decimal_test = """
@external
def phooey(inp: decimal) -> decimal:
    x: decimal = 10000.0
    for i: uint256 in range(4):
        x = x * inp
    return x

@external
def arg(inp: decimal) -> decimal:
    return inp

@external
def garg() -> decimal:
    x: decimal = 4.5
    x *= 1.5
    return x

@external
def harg() -> decimal:
    x: decimal = 4.5
    x *= 2.0
    return x

@external
def iarg() -> uint256:
    x: uint256 = as_wei_value(7, "wei")
    x *= 2
    return x
    """

    c = get_contract(harder_decimal_test)
    assert c.phooey(decimal_to_int("1.2")) == decimal_to_int("20736.0")
    assert c.phooey(decimal_to_int("-1.2")) == decimal_to_int("20736.0")
    assert c.arg(decimal_to_int("-3.7")) == decimal_to_int("-3.7")
    assert c.arg(decimal_to_int("3.7")) == decimal_to_int("3.7")
    assert c.garg() == decimal_to_int("6.75")
    assert c.harg() == decimal_to_int("9.0")
    assert c.iarg() == 14

    print("Passed fractional multiplication test")


def test_mul_overflow(tx_failed, get_contract):
    mul_code = """

@external
def _num_mul(x: decimal, y: decimal) -> decimal:
    return x * y

    """

    c = get_contract(mul_code)

    x = decimal_to_int("85070591730234615865843651857942052864")
    y = decimal_to_int("136112946768375385385349842973")

    with tx_failed():
        c._num_mul(x, y)

    x = SizeLimits.MAX_AST_DECIMAL
    y = 1 + DECIMAL_EPSILON

    with tx_failed():
        c._num_mul(decimal_to_int(x), decimal_to_int(y))

    assert c._num_mul(decimal_to_int(x), decimal_to_int(1)) == decimal_to_int(x)

    assert c._num_mul(decimal_to_int(x), decimal_to_int(1 - DECIMAL_EPSILON)) == decimal_to_int(
        quantize(x * (1 - DECIMAL_EPSILON))
    )

    x = SizeLimits.MIN_AST_DECIMAL
    assert c._num_mul(decimal_to_int(x), decimal_to_int(1 - DECIMAL_EPSILON)) == decimal_to_int(
        quantize(x * (1 - DECIMAL_EPSILON))
    )


# division failure modes(!)
def test_div_overflow(get_contract, tx_failed):
    code = """
@external
def foo(x: decimal, y: decimal) -> decimal:
    return x / y
    """

    c = get_contract(code)

    x = SizeLimits.MIN_AST_DECIMAL
    y = -DECIMAL_EPSILON

    with tx_failed():
        c.foo(decimal_to_int(x), decimal_to_int(y))
    with tx_failed():
        c.foo(decimal_to_int(x), 0)
    with tx_failed():
        c.foo(decimal_to_int(y), 0)

    y = 1 - DECIMAL_EPSILON  # 0.999999999
    with tx_failed():
        c.foo(decimal_to_int(x), decimal_to_int(y))

    y = -1
    with tx_failed():
        c.foo(decimal_to_int(x), decimal_to_int(y))

    assert c.foo(decimal_to_int(x), decimal_to_int(1)) == decimal_to_int(x)
    assert c.foo(decimal_to_int(x), decimal_to_int(1 + DECIMAL_EPSILON)) == decimal_to_int(
        quantize(x / (1 + DECIMAL_EPSILON))
    )

    x = SizeLimits.MAX_AST_DECIMAL

    with tx_failed():
        c.foo(decimal_to_int(x), decimal_to_int(DECIMAL_EPSILON))

    y = 1 - DECIMAL_EPSILON
    with tx_failed():
        c.foo(decimal_to_int(x), decimal_to_int(y))

    assert c.foo(decimal_to_int(x), decimal_to_int(1)) == decimal_to_int(x)

    assert c.foo(decimal_to_int(x), decimal_to_int(1 + DECIMAL_EPSILON)) == decimal_to_int(
        quantize(x / (1 + DECIMAL_EPSILON))
    )


def test_decimal_min_max_literals(tx_failed, get_contract):
    code = """
@external
def maximum():
    a: decimal = 18707220957835557353007165858768422651595.9365500927
@external
def minimum():
    a: decimal = -18707220957835557353007165858768422651595.9365500928
    """
    c = get_contract(code)

    assert c.maximum() is None
    assert c.minimum() is None


def test_scientific_notation(get_contract):
    code = """
@external
def foo() -> decimal:
    return 1e-10

@external
def bar(num: decimal) -> decimal:
    return num + -1e38
    """
    c = get_contract(code)

    assert c.foo() == decimal_to_int("1e-10")  # Smallest possible decimal
    assert c.bar(decimal_to_int("1e37")) == decimal_to_int("-9e37")  # Math lines up


def test_exponents():
    code = """
@external
def foo() -> decimal:
    return 2.2 ** 2.0
    """

    with pytest.raises(TypeMismatch):
        compile_code(code)


def test_decimal_nested_intermediate_overflow():
    code = """
@external
def foo():
    a: decimal = 18707220957835557353007165858768422651595.9365500927 + 1e-10 - 1e-10
    """
    with pytest.raises(OverflowException):
        compile_code(code)


def test_replace_decimal_nested_intermediate_underflow(dummy_input_bundle):
    code = """
@external
def foo():
    a: decimal = -18707220957835557353007165858768422651595.9365500928 - 1e-10 + 1e-10
    """
    with pytest.raises(OverflowException):
        compile_code(code)


def test_invalid_floordiv():
    code = """
@external
def foo():
    a: decimal = 5.0 // 9.0
    """
    with pytest.raises(InvalidOperation) as e:
        compile_code(code)

    assert e.value._hint == "did you mean `5.0 / 9.0`?"


def test_decimals_blocked():
    code = """
@external
def foo(x: decimal):
    pass
    """
    # enable_decimals default to False normally, but defaults to True in the
    # test suite. this test manually overrides the default value to test the
    # "normal" behavior of enable_decimals outside of the test suite.
    try:
        assert compiler_settings.DEFAULT_ENABLE_DECIMALS is True
        compiler_settings.DEFAULT_ENABLE_DECIMALS = False
        with pytest.raises(FeatureException) as e:
            compile_code(code)
        assert e.value._message == "decimals are not allowed unless `--enable-decimals` is set"
    finally:
        compiler_settings.DEFAULT_ENABLE_DECIMALS = True



import pytest
from eth_utils import to_bytes

from vyper import compiler
from vyper.exceptions import StructureException, TypeMismatch
from vyper.utils import method_id


def test_variable_assignment(get_contract):
    code = """
@external
def foo() -> Bytes[4]:
    bar: Bytes[4] = slice(msg.data, 0, 4)
    return bar
"""

    contract = get_contract(code)
    assert contract.foo() == method_id("foo()")


def test_slicing_start_index_other_than_zero(get_contract):
    code = """
@external
def foo(_value: uint256) -> uint256:
    bar: Bytes[32] = slice(msg.data, 4, 32)
    return convert(bar, uint256)
"""

    contract = get_contract(code)

    assert contract.foo(42) == 42


def test_get_full_calldata(get_contract):
    code = """
@external
def foo(bar: uint256) -> Bytes[36]:
    data: Bytes[36] = slice(msg.data, 0, 36)
    return data
"""
    contract = get_contract(code)

    # 2fbebd38000000000000000000000000000000000000000000000000000000000000002a
    foo_method_id = method_id("foo(uint256)")  # 2fbebd38
    encoded_42 = to_bytes(42)  # 2a
    expected_result = foo_method_id + b"\0" * 31 + encoded_42

    assert contract.foo(42) == expected_result


@pytest.mark.parametrize("bar", [0, 1, 42, 2**256 - 1])
def test_calldata_private(get_contract, bar):
    code = """
@external
def foo(bar: uint256) -> uint256:
    data: Bytes[32] = slice(msg.data, 4, 32)
    return convert(data, uint256)
    """
    c = get_contract(code)

    assert c.foo(bar) == bar


def test_memory_pointer_advances_appropriately(get_contract, keccak):
    code = """
@external
def foo() -> (uint256, Bytes[4], uint256):
    a: uint256 = max_value(uint256)
    b: Bytes[4] = slice(msg.data, 0, 4)
    c: uint256 = max_value(uint256)

    return (a, b, c)
"""
    contract = get_contract(code)

    assert contract.foo() == (2**256 - 1, method_id("foo()"), 2**256 - 1)


def test_assignment_to_storage(get_contract, keccak):
    code = """
cache: public(Bytes[4])

@external
def foo():
    self.cache = slice(msg.data, 0, 4)
"""
    contract = get_contract(code)
    contract.foo()
    assert contract.cache() == method_id("foo()")


def test_get_len(get_contract):
    code = """
@external
def foo(bar: uint256) -> uint256:
    return len(msg.data)
"""
    contract = get_contract(code)

    assert contract.foo(42) == 36


fail_list = [
    (
        """
@external
def foo() -> Bytes[4]:
    bar: Bytes[4] = msg.data
    return bar
    """,
        StructureException,
    ),
    (
        """
@external
def foo() -> Bytes[7]:
    bar: Bytes[7] = concat(msg.data, 0xc0ffee)
    return bar
    """,
        StructureException,
    ),
    (
        """
@external
def foo() -> uint256:
    bar: uint256 = convert(msg.data, uint256)
    return bar
    """,
        StructureException,
    ),
    (
        """
a: HashMap[Bytes[10], uint256]

@external
def foo():
    self.a[msg.data] += 1
    """,
        StructureException,
    ),
    (
        """
@external
def foo(bar: uint256) -> bytes32:
    ret_val: bytes32 = slice(msg.data, 4, 32)
    return ret_val
    """,
        TypeMismatch,
    ),
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_invalid_usages_compile_error(bad_code):
    with pytest.raises(bad_code[1]):
        compiler.compile_code(bad_code[0])


def test_runtime_failure_bounds_check(get_contract, tx_failed):
    code = """
@external
def foo(_value: uint256) -> uint256:
    val: Bytes[40] = slice(msg.data, 0, 40)
    return convert(slice(val, 4, 32), uint256)
"""

    contract = get_contract(code)
    with tx_failed():
        contract.foo(42)



import pytest

from vyper.exceptions import InstantiationException

invalid_list = [
    """
event Foo:
    a: uint256

@external
def foo() -> Foo:
    return Foo(2)
    """,
    """
event Foo:
    a: uint256

@external
def foo() -> (uint256, Foo):
    return 1, Foo(2)
    """,
    """
a: HashMap[uint256, uint256]

@external
def foo() -> HashMap[uint256, uint256]:
    return self.a
    """,
    """
event Foo:
    a: uint256

@external
def foo(x: Foo):
    pass
    """,
    """
@external
def foo(x: HashMap[uint256, uint256]):
    pass
    """,
    """
event Foo:
    a: uint256

foo: Foo
    """,
    """
event Foo:
    a: uint256

@external
def foo():
    f: Foo = Foo(1)
    pass
    """,
    """
event Foo:
    a: uint256

b: HashMap[uint256, Foo]
    """,
    """
event Foo:
    a: uint256

b: HashMap[Foo, uint256]
    """,
    """
b: immutable(HashMap[uint256, uint256])

@deploy
def __init__():
    b = empty(HashMap[uint256, uint256])
    """,
]


@pytest.mark.parametrize("bad_code", invalid_list)
def test_instantiation_exception(bad_code, get_contract, assert_compile_failed):
    assert_compile_failed(lambda: get_contract(bad_code), InstantiationException)



from vyper import ast as vy_ast
from vyper.exceptions import InvalidLiteral, UnfoldableNode, VyperException
from vyper.semantics.analysis.base import VarInfo
from vyper.semantics.analysis.common import VyperNodeVisitorBase
from vyper.semantics.namespace import get_namespace


def constant_fold(module_ast: vy_ast.Module):
    ConstantFolder(module_ast).run()


class ConstantFolder(VyperNodeVisitorBase):
    def __init__(self, module_ast):
        self._constants = {}
        self._module_ast = module_ast

    def run(self):
        self._get_constants()
        self.visit(self._module_ast)

    def _get_constants(self):
        module = self._module_ast
        const_var_decls = module.get_children(vy_ast.VariableDecl, {"is_constant": True})

        while True:
            n_processed = 0

            for c in const_var_decls.copy():
                # visit the entire constant node in case its type annotation
                # has unfolded constants in it.
                self.visit(c)

                assert c.value is not None  # guaranteed by VariableDecl.validate()
                try:
                    val = c.value.get_folded_value()
                except UnfoldableNode:
                    # not foldable, maybe it depends on other constants
                    # so try again later
                    continue

                # note that if a constant is redefined, its value will be
                # overwritten, but it is okay because the error is handled
                # downstream
                name = c.target.id
                self._constants[name] = val

                n_processed += 1
                const_var_decls.remove(c)

            if n_processed == 0:
                # this condition means that there are some constant vardecls
                # whose values are not foldable. this can happen for struct
                # and interface constants for instance. these are valid constant
                # declarations, but we just can't fold them at this stage.
                break

    def visit(self, node):
        if node.has_folded_value:
            return node.get_folded_value()

        for c in node.get_children():
            try:
                self.visit(c)
            except UnfoldableNode:
                # ignore bubbled up exceptions
                pass

        try:
            for class_ in node.__class__.mro():
                ast_type = class_.__name__

                visitor_fn = getattr(self, f"visit_{ast_type}", None)
                if visitor_fn:
                    folded_value = visitor_fn(node)
                    node._set_folded_value(folded_value)
                    return folded_value
        except UnfoldableNode:
            # ignore bubbled up exceptions
            pass

        return node

    def visit_Constant(self, node) -> vy_ast.ExprNode:
        return node

    def visit_Name(self, node) -> vy_ast.ExprNode:
        try:
            return self._constants[node.id]
        except KeyError:
            raise UnfoldableNode("unknown name", node)

    def visit_Attribute(self, node) -> vy_ast.ExprNode:
        namespace = get_namespace()
        path = []
        value = node.value
        while isinstance(value, vy_ast.Attribute):
            path.append(value.attr)
            value = value.value

        path.reverse()

        if not isinstance(value, vy_ast.Name):
            raise UnfoldableNode("not a module", value)

        # not super type-safe but we don't care. just catch AttributeErrors
        # and move on
        try:
            module_t = namespace[value.id].module_t

            for module_name in path:
                module_t = module_t.members[module_name].module_t

            varinfo = module_t.get_member(node.attr, node)

            return varinfo.decl_node.value.get_folded_value()
        except (VyperException, AttributeError, KeyError):
            raise UnfoldableNode("not a module")

    def visit_UnaryOp(self, node):
        operand = node.operand.get_folded_value()

        if isinstance(node.op, vy_ast.Not) and not isinstance(operand, vy_ast.NameConstant):
            raise UnfoldableNode("not a boolean!", node.operand)
        if isinstance(node.op, vy_ast.USub) and not isinstance(operand, vy_ast.Num):
            raise UnfoldableNode("not a number!", node.operand)
        if isinstance(node.op, vy_ast.Invert) and not isinstance(operand, vy_ast.Int):
            raise UnfoldableNode("not an int!", node.operand)

        value = node.op._op(operand.value)
        return type(operand).from_node(node, value=value)

    def visit_BinOp(self, node):
        left, right = [i.get_folded_value() for i in (node.left, node.right)]
        if type(left) is not type(right):
            raise UnfoldableNode("invalid operation", node)
        if not isinstance(left, vy_ast.Num):
            raise UnfoldableNode("not a number!", node.left)

        # this validation is performed to prevent the compiler from hanging
        # on very large shifts and improve the error message for negative
        # values.
        if isinstance(node.op, (vy_ast.LShift, vy_ast.RShift)) and not (0 <= right.value <= 256):
            raise InvalidLiteral("Shift bits must be between 0 and 256", node.right)

        value = node.op._op(left.value, right.value)
        return type(left).from_node(node, value=value)

    def visit_BoolOp(self, node):
        values = [v.get_folded_value() for v in node.values]

        if any(not isinstance(v, vy_ast.NameConstant) for v in values):
            raise UnfoldableNode("Node contains invalid field(s) for evaluation")

        values = [v.value for v in values]
        value = node.op._op(values)
        return vy_ast.NameConstant.from_node(node, value=value)

    def visit_Compare(self, node):
        left, right = [i.get_folded_value() for i in (node.left, node.right)]
        if not isinstance(left, vy_ast.Constant):
            raise UnfoldableNode("Node contains invalid field(s) for evaluation")

        # CMC 2022-08-04 we could probably remove these evaluation rules as they
        # are taken care of in the IR optimizer now.
        if isinstance(node.op, (vy_ast.In, vy_ast.NotIn)):
            if not isinstance(right, vy_ast.List):
                raise UnfoldableNode("Node contains invalid field(s) for evaluation")
            if next((i for i in right.elements if not isinstance(i, vy_ast.Constant)), None):
                raise UnfoldableNode("Node contains invalid field(s) for evaluation")
            if len(set([type(i) for i in right.elements])) > 1:
                raise UnfoldableNode("List contains multiple literal types")
            value = node.op._op(left.value, [i.value for i in right.elements])
            return vy_ast.NameConstant.from_node(node, value=value)

        if not isinstance(left, type(right)):
            raise UnfoldableNode("Cannot compare different literal types")

        # this is maybe just handled in the type checker.
        if not isinstance(node.op, (vy_ast.Eq, vy_ast.NotEq)) and not isinstance(left, vy_ast.Num):
            raise UnfoldableNode(
                f"Invalid literal types for {node.op.description} comparison", node
            )

        value = node.op._op(left.value, right.value)
        return vy_ast.NameConstant.from_node(node, value=value)

    def visit_List(self, node) -> vy_ast.ExprNode:
        elements = [e.get_folded_value() for e in node.elements]
        return type(node).from_node(node, elements=elements)

    def visit_Tuple(self, node) -> vy_ast.ExprNode:
        elements = [e.get_folded_value() for e in node.elements]
        return type(node).from_node(node, elements=elements)

    def visit_Dict(self, node) -> vy_ast.ExprNode:
        values = [v.get_folded_value() for v in node.values]
        return type(node).from_node(node, values=values)

    def visit_Call(self, node) -> vy_ast.ExprNode:
        if not isinstance(node.func, vy_ast.Name):
            raise UnfoldableNode("not a builtin", node)

        namespace = get_namespace()

        func_name = node.func.id
        if func_name not in namespace:
            raise UnfoldableNode("unknown", node)

        varinfo = namespace[func_name]
        if not isinstance(varinfo, VarInfo):
            raise UnfoldableNode("unfoldable", node)

        typ = varinfo.typ
        # TODO: rename to vyper_type.try_fold_call_expr
        if not hasattr(typ, "_try_fold"):
            raise UnfoldableNode("unfoldable", node)
        return typ._try_fold(node)

    def visit_Subscript(self, node) -> vy_ast.ExprNode:
        slice_ = node.slice.get_folded_value()
        value = node.value.get_folded_value()

        if not isinstance(value, vy_ast.List):
            raise UnfoldableNode("Subscript object is not a literal list")

        elements = value.elements
        if len(set([type(i) for i in elements])) > 1:
            raise UnfoldableNode("List contains multiple node types")

        if not isinstance(slice_, vy_ast.Int):
            raise UnfoldableNode("invalid index type", slice_)

        idx = slice_.value
        if idx < 0 or idx >= len(elements):
            raise UnfoldableNode("invalid index value")

        return elements[idx]



"""
isort:skip_file
"""
import sys

from . import nodes, validation
from .natspec import parse_natspec
from .nodes import as_tuple
from .utils import ast_to_dict
from .parse import parse_to_ast, parse_to_ast_with_settings

# adds vyper.ast.nodes classes into the local namespace
for name, obj in (
    (k, v) for k, v in nodes.__dict__.items() if type(v) is type and nodes.VyperNode in v.__mro__
):
    setattr(sys.modules[__name__], name, obj)



from vyper.venom.analysis.cfg import CFGAnalysis
from vyper.venom.analysis.dfg import DFGAnalysis
from vyper.venom.analysis.dominators import DominatorTreeAnalysis
from vyper.venom.analysis.liveness import LivenessAnalysis
from vyper.venom.basicblock import IRVariable
from vyper.venom.passes.base_pass import IRPass


class StoreElimination(IRPass):
    """
    This pass forwards variables to their uses though `store` instructions,
    and removes the `store` instruction.
    """

    def run_pass(self):
        self.analyses_cache.request_analysis(CFGAnalysis)
        dfg = self.analyses_cache.request_analysis(DFGAnalysis)

        for var, inst in dfg.outputs.items():
            if inst.opcode != "store":
                continue
            if not isinstance(inst.operands[0], IRVariable):
                continue
            if inst.operands[0].name in ["%ret_ofst", "%ret_size"]:
                continue
            if inst.output.name in ["%ret_ofst", "%ret_size"]:
                continue
            self._process_store(dfg, inst, var, inst.operands[0])

        self.analyses_cache.invalidate_analysis(DominatorTreeAnalysis)
        self.analyses_cache.invalidate_analysis(LivenessAnalysis)
        self.analyses_cache.invalidate_analysis(DFGAnalysis)

    def _process_store(self, dfg, inst, var, new_var):
        """
        Process store instruction. If the variable is only used by a load instruction,
        forward the variable to the load instruction.
        """
        uses = dfg.get_uses(var)

        if any([inst.opcode == "phi" for inst in uses]):
            return

        for use_inst in uses:
            for i, operand in enumerate(use_inst.operands):
                if operand == var:
                    use_inst.operands[i] = new_var

        inst.parent.remove_instruction(inst)



import pytest

from vyper.compiler import compile_code
from vyper.compiler.settings import OptimizationLevel, Settings
from vyper.exceptions import SyntaxException


def test_semicolon_prohibited(get_contract):
    code = """@external
def test() -> int128:
    a: int128 = 1; b: int128 = 2
    return a + b
    """

    with pytest.raises(SyntaxException):
        get_contract(code)


def test_valid_semicolons(get_contract):
    code = """
@external
def test() -> int128:
    a: int128 = 1
    b: int128 = 2
    s: String[300] = "this should not be a problem; because it is in a string"
    s = \"\"\"this should not be a problem; because it's in a string\"\"\"
    s = 'this should not be a problem;;; because it\\\'s in a string'
    s = '''this should not ; \'cause it\'s in a string'''
    s = "this should not be \\\"; because it's in a ;\\\"string;\\\";"
    return a + b
    """
    c = get_contract(code)
    assert c.test() == 3


def test_external_contract_definition_alias(get_contract):
    contract_1 = """
@external
def bar() -> int128:
    return 1
    """

    contract_2 = """
interface Bar:
    def bar() -> int128: nonpayable

bar_contract: Bar

@external
def foo(contract_address: address) -> int128:
    self.bar_contract = Bar(contract_address)
    return extcall self.bar_contract.bar()
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)
    assert c2.foo(c1.address) == 1


def test_version_pragma(get_contract):
    from vyper import __version__

    installed_version = ".".join(__version__.split(".")[:3])

    code = f"""
# @version {installed_version}

@external
def test():
    pass
    """
    assert get_contract(code)


def test_version_pragma2(get_contract):
    # new, `#pragma` way of doing things
    from vyper import __version__

    installed_version = ".".join(__version__.split(".")[:3])

    code = f"""
#pragma version {installed_version}

@external
def test():
    pass
    """
    assert get_contract(code)


def test_evm_version_check(assert_compile_failed):
    code = """
#pragma evm-version london
    """
    assert compile_code(code, settings=Settings(evm_version=None)) is not None
    assert compile_code(code, settings=Settings(evm_version="london")) is not None
    # should fail if compile options indicate different evm version
    # from source pragma
    with pytest.raises(ValueError):
        compile_code(code, settings=Settings(evm_version="shanghai"))


def test_optimization_mode_check():
    code = """
#pragma optimize codesize
    """
    assert compile_code(code, settings=Settings(optimize=None))
    # should fail if compile options indicate different optimization mode
    # from source pragma
    with pytest.raises(ValueError):
        compile_code(code, settings=Settings(optimize=OptimizationLevel.GAS))
    with pytest.raises(ValueError):
        compile_code(code, settings=Settings(optimize=OptimizationLevel.NONE))


def test_optimization_mode_check_none():
    code = """
#pragma optimize none
    """
    assert compile_code(code, settings=Settings(optimize=None))
    # "none" conflicts with "gas"
    with pytest.raises(ValueError):
        compile_code(code, settings=Settings(optimize=OptimizationLevel.GAS))


def test_version_empty_version(assert_compile_failed, get_contract):
    code = """
#@version

@external
def test():
    pass
    """
    assert_compile_failed(lambda: get_contract(code))


def test_version_empty_version_mismatch(assert_compile_failed, get_contract):
    code = """
# @version 9.9.9

@external
def test():
    pass
    """
    assert_compile_failed(lambda: get_contract(code))


def test_version_empty_invalid_version_string(assert_compile_failed, get_contract):
    code = """
# @version hello

@external
def test():
    pass
    """
    assert_compile_failed(lambda: get_contract(code))


def test_unbalanced_parens(assert_compile_failed, get_contract):
    code = """
@external
def foo():
    convert(
    """

    with pytest.raises(SyntaxException):
        get_contract(code)



import contextlib

from vyper.ast.identifiers import validate_identifier
from vyper.exceptions import CompilerPanic, NamespaceCollision, UndeclaredDefinition
from vyper.semantics.analysis.levenshtein_utils import get_levenshtein_error_suggestions


class Namespace(dict):
    """
    Dictionary subclass that represents the namespace of a contract.

    Attributes
    ----------
    _scopes : List[Set]
        List of sets containing the key names for each scope
    """

    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls, *args, **kwargs)
        self._scopes = []
        return self

    def __init__(self):
        super().__init__()
        # NOTE cyclic imports!
        # TODO: break this cycle by providing an `init_vyper_namespace` in 3rd module
        from vyper.builtins.functions import get_builtin_functions
        from vyper.semantics import environment
        from vyper.semantics.analysis.base import VarInfo
        from vyper.semantics.types import PRIMITIVE_TYPES

        self.update(PRIMITIVE_TYPES)
        self.update(environment.get_constant_vars())
        self.update({k: VarInfo(b) for (k, b) in get_builtin_functions().items()})

    def __eq__(self, other):
        return self is other

    def __setitem__(self, attr, obj):
        if self._scopes:
            self.validate_assignment(attr)
            self._scopes[-1].add(attr)
        assert isinstance(attr, str), f"not a string: {attr}"
        super().__setitem__(attr, obj)

    def __getitem__(self, key):
        if key not in self:
            hint = get_levenshtein_error_suggestions(key, self, 0.2)
            raise UndeclaredDefinition(f"'{key}' has not been declared.", hint=hint)
        return super().__getitem__(key)

    def __enter__(self):
        if not self._scopes:
            raise CompilerPanic("Context manager must be invoked via namespace.enter_scope()")

    def __exit__(self, exc_type, exc_value, traceback):
        if not self._scopes:
            raise CompilerPanic("Bad use of namespace as a context manager")
        for key in self._scopes.pop():
            del self[key]

    def enter_scope(self):
        """
        Enter a new scope within the namespace.

        Called as a context manager, e.g. `with namespace.enter_scope():`
        All items that are added within the context are removed upon exit.
        """
        # NOTE cyclic imports!
        from vyper.semantics import environment

        self._scopes.append(set())

        if len(self._scopes) == 1:
            # add mutable vars (`self`) to the initial scope
            self.update(environment.get_mutable_vars())

        return self

    def update(self, other):
        for key, value in other.items():
            self.__setitem__(key, value)

    def clear(self):
        super().clear()
        self.__init__()

    def validate_assignment(self, attr):
        validate_identifier(attr)

        if attr in self:
            prev = super().__getitem__(attr)
            prev_decl = getattr(prev, "decl_node", None)
            msg = f"'{attr}' has already been declared"
            if prev_decl is None:
                msg += f" as a {prev}"
            raise NamespaceCollision(msg, prev_decl=prev_decl)


def get_namespace():
    """
    Get the global namespace object.
    """
    global _namespace
    try:
        return _namespace
    except NameError:
        _namespace = Namespace()
        return _namespace


@contextlib.contextmanager
def override_global_namespace(ns):
    global _namespace
    tmp = _namespace
    try:
        # clobber global namespace
        _namespace = ns
        yield
    finally:
        # unclobber
        _namespace = tmp



import pytest
from eth_account import Account
from eth_account.messages import encode_defunct
from eth_keys import KeyAPI
from eth_utils import is_same_address, to_bytes, to_checksum_address, to_int

from tests.utils import ZERO_ADDRESS
from vyper.utils import keccak256


@pytest.fixture(scope="module")
def c(env, get_contract):
    a0, a1, a2, a3, a4, a5, a6 = env.accounts[:7]
    with open("examples/wallet/wallet.vy") as f:
        code = f.read()
    # Sends wei to the contract for future transactions gas costs
    c = get_contract(code, *[[a1, a2, a3, a4, a5], 3])
    env.set_balance(a0, 10**17)
    env.message_call(c.address, value=10**17)
    return c


def sign(seq, to, value, data, key):
    keys = KeyAPI()
    comb = seq.to_bytes(32, "big") + b"\x00" * 12 + to + value.to_bytes(32, "big") + data
    h1 = keccak256(comb)
    h2 = keccak256(b"\x19Ethereum Signed Message:\n32" + h1)
    sig = keys.ecdsa_sign(h2, key)
    return [28 if sig.v == 1 else 27, sig.r, sig.s]


def test_approve(env, c, tx_failed):
    a0, a1, a2, a3, a4, a5, a6 = env.accounts[:7]
    k0, k1, k2, k3, k4, k5, k6, k7 = env._keys[:8]
    env.set_balance(a1, 10**18)

    to, value, data = b"\x35" * 20, 10**16, b""
    to_address = to_checksum_address(to)

    def pack_and_sign(seq, *args):
        sigs = [sign(seq, to, value, data, k) if k else [0, 0, 0] for k in args]
        return sigs

    # Legitimate approval
    sigs = pack_and_sign(0, k1, 0, k3, 0, k5)
    c.approve(0, "0x" + to.hex(), value, data, sigs, value=value, sender=a1)
    # Approve fails if only 2 signatures are given
    sigs = pack_and_sign(1, k1, 0, k3, 0, 0)
    with tx_failed():
        c.approve(1, to_address, value, data, sigs, value=value, sender=a1)
    # Approve fails if an invalid signature is given
    sigs = pack_and_sign(1, k1, 0, k7, 0, k5)
    with tx_failed():
        c.approve(1, to_address, value, data, sigs, value=value, sender=a1)
    # Approve fails if transaction number is incorrect (the first argument should be 1)
    sigs = pack_and_sign(0, k1, 0, k3, 0, k5)
    with tx_failed():
        c.approve(0, to_address, value, data, sigs, value=value, sender=a1)
    # Approve fails if not enough value is sent
    sigs = pack_and_sign(1, k1, 0, k3, 0, k5)
    with tx_failed():
        c.approve(1, to_address, value, data, sigs, value=0, sender=a1)
    sigs = pack_and_sign(1, k1, 0, k3, 0, k5)

    # this call should succeed
    c.approve(1, to_address, value, data, sigs, value=value, sender=a1)

    print("Basic tests passed")


def test_javascript_signatures(env, get_contract, keccak):
    a3 = env.accounts[2]
    # The zero address will cause `approve` to default to valid signatures
    accounts = [
        "0x776ba14735ff84789320718cf0aa43e91f7a8ce1",
        "0x095ce4e4240fa66ff90282c26847456e3f3b5002",
    ]
    # The address that will receive the transaction
    recipient = "0x776Ba14735FF84789320718cf0aa43e91F7A8Ce1"
    # These are the matching sigs to the accounts
    raw_sigs = [
        "0x4a89507bf71749fb338ed13fba623a683d9ecab0fb9c389a4298525c043e38281a00ab65628bb18a382eb8c8b4fb4dae95ccc993cf49f617c60d8051180778601c",  # noqa: E501
        "0xc84fe5d2a600e033930e0cf73f26e78f4c65b134f9c9992f60f08ce0863abdbe0548a6e8aa2d952659f29c67106b59fdfcd64d67df03c1df620c70c85578ae701b",  # noqa: E501
    ]

    # Turns the raw sigs into sigs
    sigs = [
        (to_int(x[64:]), to_int(x[:32]), to_int(x[32:64]))  # v  # r  # s
        for x in map(lambda z: to_bytes(hexstr=z[2:]), raw_sigs)
    ]

    h = keccak(
        (0).to_bytes(32, "big")
        + b"\x00" * 12
        + to_bytes(hexstr=recipient[2:])
        + (25).to_bytes(32, "big")
        + b""
    )  # noqa: E501
    h2 = encode_defunct(h)

    # Check to make sure the signatures are valid
    assert is_same_address(Account.recover_message(h2, sigs[0]), accounts[0])
    assert is_same_address(Account.recover_message(h2, sigs[1]), accounts[1])

    # Set the owners to zero addresses
    with open("examples/wallet/wallet.vy") as f:
        owners = [to_checksum_address(x) for x in accounts + [a3, ZERO_ADDRESS, ZERO_ADDRESS]]
        x2 = get_contract(f.read(), *[owners, 2])

    env.set_balance(env.deployer, 10**18)
    env.message_call(x2.address, value=10**17)

    # There's no need to pass in signatures because the owners are 0 addresses
    # causing them to default to valid signatures
    x2.approve(0, recipient, 25, b"", sigs + [[0, 0, 0]] * 3, value=10**17)



def test_method_id_test(get_contract):
    method_id_test = """
@external
def double(x: int128) -> int128:
    return x * 2

@external
def returnten() -> int128:
    ans: Bytes[32] = raw_call(self, concat(method_id("double(int128)"), convert(5, bytes32)), gas=50000, max_outsize=32)  # noqa: E501
    return convert(convert(ans, bytes32), int128)
    """
    c = get_contract(method_id_test)
    assert c.returnten() == 10


def test_method_id_bytes4(get_contract):
    code = """
@external
def sig() -> bytes4:
    return method_id('transfer(address,uint256)', output_type=bytes4)
    """
    c = get_contract(code)
    sig = c.sig()

    assert sig == b"\xa9\x05\x9c\xbb"


def test_method_id_Bytes4(get_contract):
    code = """
@external
def sig() -> Bytes[4]:
    return method_id('transfer(address,uint256)', output_type=Bytes[4])
    """
    c = get_contract(code)
    sig = c.sig()

    # assert len(sig) == 4
    assert sig == b"\xa9\x05\x9c\xbb"


def test_method_id_bytes4_default(get_contract):
    code = """
@external
def sig() -> Bytes[4]:
    return method_id('transfer(address,uint256)')
    """
    c = get_contract(code)
    sig = c.sig()

    # assert len(sig) == 4
    assert sig == b"\xa9\x05\x9c\xbb"


def test_method_id_invalid_space(get_contract, assert_compile_failed):
    code = """
@external
def sig() -> bytes4:
    return method_id('transfer(address, uint256)', output_type=bytes4)
    """
    assert_compile_failed(lambda: get_contract(code))


def test_method_id_invalid_type(get_contract, assert_compile_failed):
    code = """
@external
def sig() -> bytes32:
    return method_id('transfer(address,uint256)', output_type=bytes32)
    """
    assert_compile_failed(lambda: get_contract(code))



def test_null_code(get_contract):
    null_code = """
@external
def foo():
    pass
    """
    c = get_contract(null_code)
    c.foo()


def test_basic_code(get_contract):
    basic_code = """
@external
def foo(x: int128) -> int128:
    return x * 2

    """
    c = get_contract(basic_code)
    assert c.foo(9) == 18



import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from tests.utils import parse_and_fold

st_uint256 = st.integers(min_value=0, max_value=(2**256 - 1))


@pytest.mark.fuzzing
@settings(max_examples=100)
@given(a=st_uint256, b=st_uint256)
def test_powmod_uint256(get_contract, a, b):
    source = """
@external
def foo(a: uint256, b: uint256) -> uint256:
    return pow_mod256(a, b)
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"pow_mod256({a}, {b})")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(a, b) == new_node.value



import pytest
from pytest import raises

from vyper import compiler
from vyper.exceptions import CallViolation

call_violation_list = [
    """
f:int128

@external
def a (x:int128)->int128:
    self.f = 100
    return x+5

@view
@external
def b():
    p: int128 = self.a(10)
    """,
    """
@external
def goo():
    pass

@internal
def foo():
    self.goo()
    """,
    """
@deploy
def __init__():
    pass

@internal
def foo():
    self.__init__()
    """,
]


@pytest.mark.parametrize("bad_code", call_violation_list)
def test_call_violation_exception(bad_code):
    with raises(CallViolation):
        compiler.compile_code(bad_code)
